from cobaya.likelihoods.base_classes import BAO

"""
# DESI BAO likelihoods

## version 1.0.0

BAO likelihoods corresponding to the arXiv release on April 5, 2024, see https://arxiv.org/abs/2404.03000, https://arxiv.org/abs/2404.03001, https://arxiv.org/abs/2404.03002.
**desi_2024_bao_all** contains all DESI BAO measurements, and should be used as a baseline.
BAO measurements in each redshift bin (that are considered independent) are also provided:
- **desi_2024_bao_bgs_z1**: BGS, 0.1 < z < 0.4
- **desi_2024_bao_lrg_z1**: LRG, 0.4 < z < 0.6
- **desi_2024_bao_lrg_z2**: LRG, 0.6 < z < 0.8
- **desi_2024_bao_lrgpluselg_z1**: LRG+ELG, 0.8 < z < 1.1
- **desi_2024_bao_elg_z2**: ELG, 1.1 < z < 1.6
- **desi_2024_bao_qso_z1**: QSO, 0.8 < z < 2.1
- **desi_2024_bao_lya**: Lya

Finally, the combined SDSS(eBOSS)+DESI Lya measurement is:
- **desi_2024_eboss_bao_lya**: Lya """


class desi_2024_base(BAO):
    bibtex_file = 'desi_2024_bao.bibtex'


class desi_2024_bao_all(desi_2024_base):
    r"""
    DESI BAO likelihood for all tracers.
    """



info = {
    'params': {
        # Fixed
        'ombh2': 0.022, 'omch2': 0.12, 'H0': 68, 'tau': 0.07,
        'mnu': 0.06, 'nnu': 3.046,
        # Sampled
        'As': {'prior': {'min': 1e-9, 'max': 4e-9}, 'latex': 'A_s'},
        'ns': {'prior': {'min': 0.9, 'max': 1.1}, 'latex': 'n_s'},
        'noise_std_pixel': {
            'prior': {'dist': 'norm', 'loc': 20, 'scale': 5},
            'latex': r'\sigma_\mathrm{pix}'},
        # Derived
        'Map_Cl_at_500': {'latex': r'C_{500,\,\mathrm{map}}'}},
    'likelihood': {'my_cl_like': {
        "external": my_like,
        # Declare required quantities!
        "requires": {'Cl': {'tt': l_max}},
        # Declare derived parameters!
        "output_params": ['Map_Cl_at_500']}},
    'theory': {'camb': {'stop_at_error': True}},
    'packages_path': packages_path}

from cobaya.model import get_model
model = get_model(info)

# Eval likelihood once with fid values and plot
_do_plot = True
_plot_name = "fiducial.png"
fiducial_params_w_noise = fiducial_params.copy()
fiducial_params_w_noise['noise_std_pixel'] = 20
model.logposterior(fiducial_params_w_noise)
_do_plot = False

# Plot of (prpto) probability density
As = np.linspace(1e-9, 4e-9, 10)
loglikes = [model.loglike({'As': A, 'ns': 0.96, 'noise_std_pixel': 20})[0] for A in As]
plt.figure()
plt.plot(As, loglikes)
plt.gca().get_yaxis().set_visible(False)
plt.title(r"$\log P(A_s|\mathcal{D},\mathcal{M}) (+ \mathrm{const})$")
plt.xlabel(r"$A_s$")
plt.savefig("log_like.png")
plt.close()



import numpy as np
from scipy import stats


def gauss_ring_logp(x, y, mean_radius=1, std=0.02):
    """
    Defines a gaussian ring likelihood in cartesian coordinates,
    around some ``mean_radius`` and with some ``std``.
    """
    return stats.norm.logpdf(np.sqrt(x**2 + y**2), loc=mean_radius, scale=std)



fiducial_params = {
    'ombh2': 0.022, 'omch2': 0.12, 'H0': 68, 'tau': 0.07,
    'As': 2.2e-9, 'ns': 0.96,
    'mnu': 0.06, 'nnu': 3.046}

l_max = 1000

packages_path = '/path/to/your/packages'

info_fiducial = {
    'params': fiducial_params,
    'likelihood': {'one': None},
    'theory': {'camb': {"extra_args": {"num_massive_neutrinos": 1}}},
    'packages_path': packages_path}

from cobaya.model import get_model
model_fiducial = get_model(info_fiducial)

# Declare our desired theory product
# (there is no cosmological likelihood doing it for us)
model_fiducial.add_requirements({"Cl": {'tt': l_max}})

# Compute and extract the CMB power spectrum
# (In muK^-2, without l(l+1)/(2pi) factor)
# notice the empty dictionary below: all parameters are fixed
model_fiducial.logposterior({})
Cls = model_fiducial.provider.get_Cl(ell_factor=False, units="muK2")

# Our fiducial power spectrum
Cl_est = Cls['tt'][:l_max + 1]



Cls = model.provider.get_Cl(ell_factor=True)
import matplotlib.pyplot as plt

plt.figure(figsize=(8, 6))
plt.plot(Cls["ell"][2:], Cls["tt"][2:])
plt.ylabel(r"$\ell(\ell+1)/(2\pi)\,C_\ell\;(\mu \mathrm{K}^2)$")
plt.xlabel(r"$\ell$")
plt.savefig("cltt.png")
# plt.show()



import subprocess
import os
import numpy as np
import re
import pickle
import time
import shutil
import multiprocessing
from typing import Any

from cobaya.conventions import Extension
from .conventions import script_folder, script_ext, jobid_ext

code_prefix = 'COBAYA'
default_program = 'cobaya-run -r'


def set_default_program(program, env_prefix):
    global code_prefix, default_program
    code_prefix = env_prefix
    default_program = program


def addArguments(parser, combinedJobs=False):
    parser.add_argument('--nodes', type=int)
    parser.add_argument('--chains-per-node', type=int)
    parser.add_argument('--cores-per-node', type=int)
    parser.add_argument('--mem-per-node', type=int, help="Memory in MB per node")
    parser.add_argument('--walltime')
    if combinedJobs:
        parser.add_argument('--combine-one-job-name',
                            help=('run all one after another, under one job submission '
                                  '(good for many fast operations)'))
        parser.add_argument('--runs-per-job', type=int,
                            default=int(os.environ.get(code_prefix + '_runsPerJob', '1')),
                            help=('submit multiple mpi runs at once from each job script '
                                  '(e.g. to get more than one run per node)'))
    parser.add_argument('--job-template',
                        help="template file for the job submission script")
    parser.add_argument('--program',
                        help='actual program to run (default: %s)' % default_program)
    parser.add_argument('--queue', help='name of queue to submit to')
    parser.add_argument('--jobclass', help='any class name of the job')
    parser.add_argument('--qsub', help='option to change qsub command to something else')
    parser.add_argument('--dryrun', action='store_true',
                        help="just test configuration and give summary for checking, "
                             "don't produce or do anything")
    parser.add_argument('--no_sub', action='store_true',
                        help="produce job script but don't actually submit it")


def check_arguments(**kwargs):
    submitJob(None, None, msg=True, **kwargs)


grid_engine_defaults = {
    'PBS': {},
    'MOAB': {'qsub': 'msub', 'qdel': 'canceljob'},
    'SLURM': {'qsub': 'sbatch', 'qdel': 'scancel', 'qstat': 'squeue'}
}


def engine_default(engine, command):
    options = grid_engine_defaults.get(engine)
    if options:
        return options.get(command, command)
    return command


class JobSettings:
    names: list
    jobId: str
    path: str
    onerun: int
    inputFiles: list
    subTime: float

    def __init__(self, jobName, msg=False, **kwargs):
        self.jobName = jobName
        grid_engine = 'PBS'
        for engine, options in grid_engine_defaults.items():
            if 'qsub' in options:
                if shutil.which(options["qsub"]) is not None:
                    grid_engine = engine
                    break
        else:
            if shutil.which("qstat") is not None:
                try:
                    help_info = subprocess.check_output('qstat -help',
                                                        shell=True).decode().strip()
                    if 'OGS/GE' in help_info:
                        grid_engine = 'OGS'  # Open Grid Scheduler, as on StarCluster
                except:
                    pass
        self.job_template = get_defaulted('job_template', 'job_script', **kwargs)
        if not self.job_template or self.job_template == 'job_script' \
                and not os.path.exists(self.job_template):
            raise ValueError("You must provide a script template with '--job-template', "
                             "or export COBAYA_job_template in your .bashrc")
        try:
            with open(self.job_template, 'r', encoding="utf-8-sig") as f:
                template = f.read()
        except IOError:
            raise ValueError("Job template '%s' not found." % self.job_template)
        try:
            cores = multiprocessing.cpu_count()
            if cores > 64:
                # probably shared memory machine, e.g. Cosmos
                cores = 8
        except:
            cores = 8
        self.cores_per_node = get_defaulted('cores_per_node', cores, tp=int,
                                            template=template, **kwargs)
        if cores == 4:
            perNode = 2
        elif cores % 4 == 0:
            perNode = cores // 4
        elif cores > 3 and cores % 3 == 0:
            perNode = cores // 3
        else:
            perNode = 1
        self.chains_per_node = get_defaulted('chains_per_node', perNode, tp=int,
                                             template=template, **kwargs)
        self.nodes = get_defaulted('nodes', max(1, 4 // perNode), tp=int,
                                   template=template, **kwargs)
        self.nchains = self.nodes * self.chains_per_node
        self.runsPerJob = get_defaulted('runsPerJob', 1, tp=int, template=template,
                                        **kwargs)
        # also defaulted at input so should be set here unless called programmatically
        self.omp = self.cores_per_node / (self.chains_per_node * self.runsPerJob)
        if self.omp != np.floor(self.omp):
            raise Exception('Chains must each have equal number of cores')
        if msg:
            print('Job parameters: %i runs of %i chains on %i nodes, '
                  'each node with %i MPI chains, each chain using %i OpenMP cores '
                  '(%i cores per node)' % (
                      self.runsPerJob, self.nchains, self.nodes, self.chains_per_node,
                      self.omp, self.cores_per_node))
        self.mem_per_node = get_defaulted('mem_per_node', 63900, tp=int,
                                          template=template, **kwargs)
        self.walltime = get_defaulted('walltime', '24:00:00', template=template, **kwargs)
        self.program = get_defaulted('program', default_program, template=template,
                                     **kwargs)
        self.queue = get_defaulted('queue', '', template=template, **kwargs)
        self.jobclass = get_defaulted('jobclass', '', template=template, **kwargs)
        self.gridEngine = get_defaulted('GridEngine', grid_engine, template=template,
                                        **kwargs)
        if grid_engine == 'OGS' and os.getenv('SGE_CLUSTER_NAME', '') == 'starcluster':
            self.qsub = 'qsub -pe orte ##NUMSLOTS##'
        else:
            self.qsub = get_defaulted('qsub', engine_default(grid_engine, 'qsub'),
                                      template=template, **kwargs)
        self.qdel = get_defaulted('qdel', engine_default(grid_engine, 'qdel'),
                                  template=template, **kwargs)
        self.runCommand = extract_value(template, 'RUN')


class JobIndex:
    """
    Stores the mappings between job Ids, jobNames
    """

    def __init__(self):
        self.jobSettings = dict()
        self.jobNames = dict()
        self.rootNames = dict()
        self.jobSequence = []

    def addJob(self, j):
        self.jobSettings[j.jobId] = j
        self.jobNames[j.jobName] = j.jobId
        for name in j.names:
            self.rootNames[name] = j.jobId
        self.jobSequence.append(j.jobId)

    def delId(self, jobId):
        if jobId is not None:
            j = self.jobSettings.get(jobId)
            if j is not None:
                for rootname in j.names:
                    del self.rootNames[rootname]
                del self.jobSettings[jobId]
                del self.jobNames[j.jobName]
                self.jobSequence = [s for s in self.jobSequence if s != jobId]


def loadJobIndex(batchPath, must_exist=False):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    fileName = os.path.join(batchPath, 'job_index.pyobj')
    if os.path.exists(fileName):
        with open(fileName, 'rb') as inp:
            return pickle.load(inp)
    else:
        if not must_exist:
            return JobIndex()
        return None


def saveJobIndex(obj, batchPath=None):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    fname = os.path.join(batchPath, 'job_index.pyobj')
    with open(fname + '_tmp', 'wb') as output:
        pickle.dump(obj, output, pickle.HIGHEST_PROTOCOL)
    # try to prevent corruption from error mid-write
    if os.path.exists(fname):
        os.remove(fname)
    shutil.move(fname + '_tmp', fname)


def addJobIndex(batchPath, j):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    index.addJob(j)
    saveJobIndex(index, batchPath)


def deleteJobNames(batchPath, jobNames):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    if not index:
        raise Exception('No existing job index found')
    if isinstance(jobNames, str):
        jobNames = [jobNames]
    for name in jobNames:
        jobId = index.jobNames.get(name)
        index.delId(jobId)
    saveJobIndex(index, batchPath)


def deleteRootNames(batchPath, rootNames):
    deleteJobs(batchPath, rootNames=rootNames)


def deleteJobs(batchPath, jobIds=None, rootNames=None, jobNames=None, jobId_minmax=None,
               jobId_min=None, confirm=True, running=False, queued=False):
    if batchPath is None:
        batchPath = os.path.join(".", script_folder)
    index = loadJobIndex(batchPath)
    if not index:
        raise Exception('No existing job index found')
    if jobIds is None:
        jobIds = []
    if isinstance(jobIds, str):
        jobIds = [jobIds]
    jobIds = set(jobIds)
    if rootNames is not None:
        if isinstance(rootNames, str):
            rootNames = [rootNames]
        jobIds.update(index.rootNames.get(name) for name in rootNames)
        jobIds.discard(None)
    if jobNames is not None:
        if isinstance(jobNames, str):
            jobNames = [jobNames]
        jobIds.update(index.jobNames.get(name) for name in jobNames)
    if jobId_minmax is not None or jobId_min is not None:
        for jobIdStr, j in list(index.jobSettings.items()):
            parts = jobIdStr.split('.')
            if len(parts) == 1 or parts[0].isdigit():
                jobId = int(parts[0])
            else:
                jobId = int(parts[1])
            if (jobId_minmax is not None and
                    (jobId_minmax[0] <= jobId <= jobId_minmax[1]) or
                    jobId_min is not None and jobId >= jobId_min):
                jobIds.add(jobIdStr)
    validIds = queue_job_details(batchPath, running=not queued, queued=not running)[0]
    for jobId in sorted(jobIds):
        j = index.jobSettings.get(jobId)
        if j is not None:
            if confirm:
                if jobId in validIds:
                    print('Cancelling: ', j.jobName, jobId)
                    if hasattr(j, 'qdel'):
                        qdel = j.qdel
                    else:
                        qdel = 'qdel'
                    subprocess.check_output(qdel + ' ' + str(jobId), shell=True)
                index.delId(jobId)
            elif jobId in validIds:
                print('...', j.jobName, jobId)
    if confirm:
        saveJobIndex(index, batchPath)
    return jobIds


def submitJob(job_name, input_files, sequential=False, msg=False, **kwargs):
    """
    Submits a job with name ``jobName`` and input file(s) ``input_files``.

    If ``sequential=True`` (default: False), and multiple input files given,
    gathers the corresponding runs in a single job script, so that they are run
    sequentially.

    The ``JobSettings`` are created using the non-explicit ``kwargs``.
    """
    j = JobSettings(job_name, msg, **kwargs)
    if kwargs.get('dryrun', False) or input_files is None:
        return
    if isinstance(input_files, str):
        input_files = [input_files]
    input_files = [
        (os.path.splitext(p)[0] if os.path.splitext(p)[1] in Extension.yamls else p)
        for p in input_files]
    j.runsPerJob = (len(input_files), 1)[sequential]
    # adjust omp for the actual number
    # (may not be equal to input runsPerJob because of non-integer multiple)
    j.omp = j.cores_per_node // (j.chains_per_node * j.runsPerJob)
    j.path = os.getcwd()
    j.onerun = (0, 1)[len(input_files) == 1 or sequential]
    base_path = os.path.abspath(kwargs.get('batchPath', './'))
    script_dir = os.path.join(base_path, script_folder)
    if not os.path.exists(script_dir):
        os.makedirs(script_dir)
    vals = dict()
    vals['JOBNAME'] = job_name
    vals['OMP'] = j.omp
    vals['MEM_MB'] = j.mem_per_node
    vals['WALLTIME'] = j.walltime
    vals['NUMNODES'] = j.nodes
    vals['NUMRUNS'] = j.runsPerJob
    vals['NUMMPI'] = j.nchains
    vals['CHAINSPERNODE'] = j.chains_per_node
    vals['PPN'] = j.chains_per_node * j.runsPerJob * j.omp
    vals['MPIPERNODE'] = j.chains_per_node * j.runsPerJob
    vals['NUMTASKS'] = j.nchains * j.runsPerJob
    vals['NUMSLOTS'] = vals['PPN'] * j.nodes
    vals['ROOTDIR'] = j.path
    vals['ONERUN'] = j.onerun
    vals['PROGRAM'] = j.program
    vals['QUEUE'] = j.queue
    vals['JOBSCRIPTDIR'] = script_dir
    if hasattr(j, 'jobclass'):
        vals['JOBCLASS'] = j.jobclass
    j.names = [os.path.basename(param) for param in input_files]
    commands = []
    for param, name in zip(input_files, j.names):
        ini = param + '.yaml'
        if j.runCommand is not None:
            vals['INI'] = ini
            vals['INIBASE'] = name
            command = replace_placeholders(j.runCommand, vals)
        else:
            command = ('mpirun -np %i %s %s' % (j.nchains, j.program, ini))
        commands.append(command)
    vals['COMMAND'] = "\n".join(commands)
    with open(j.job_template, 'r', encoding="utf-8-sig") as f:
        template = f.read()
        # Remove definition lines
        template = "\n".join(
            [line for line in template.split("\n") if not line.startswith("##")])
        script = replace_placeholders(template, vals)
        scriptRoot = os.path.join(script_dir, job_name)
        scriptName = scriptRoot + script_ext
        open(scriptName, 'w', encoding="utf-8").write(script)
        if len(input_files) > 1:
            open(scriptRoot + '.batch', 'w', encoding="utf-8").write(
                "\n".join(input_files))
        if not kwargs.get('no_sub', False):
            try:
                res = subprocess.check_output(
                    replace_placeholders(j.qsub, vals) + ' ' + scriptName, shell=True,
                    stderr=subprocess.STDOUT).decode().strip()
            except subprocess.CalledProcessError as e:
                print('Error calling %s: %s' % (j.qsub, e.output.decode().strip()))
            else:
                if not res:
                    print('No qsub output')
                else:
                    j.inputFiles = input_files
                    if 'Your job ' in res:
                        m = re.search(r'Your job (\d*) ', res)
                        assert m
                        res = m.group(1)
                    j.jobId = res
                    j.subTime = time.time()
                    open(scriptRoot + jobid_ext, 'w', encoding="utf-8").write(res)
                    addJobIndex(kwargs.get('batchPath'), j)


def queue_job_details(batchPath=None, running=True, queued=True, warnNotBatch=True):
    """
    Return: list of jobIds, list of job_names, list of list names
    """
    index = loadJobIndex(batchPath)
    if not index:
        print('No existing job index found')
        return []
    if shutil.which("showq") is not None:
        qstat = 'showq'
        running_txt = ' Running '
    else:
        if shutil.which("squeue") is not None:
            qstat = 'squeue'
        else:
            # e.g. Sun Grid Engine/OGS
            qstat = 'qstat'
        running_txt = ' r '
    res_str = subprocess.check_output('%s -u $USER' % qstat, shell=True).decode().strip()
    res = res_str.split("\n")
    names = []
    job_names = []
    ids = []
    infos = []
    for line in res[2:]:
        if ((' ' + str(os.environ.get('USER')) + ' ' in line and
             (queued and not re.search(running_txt, line, re.IGNORECASE) or
              running and re.search(running_txt, line, re.IGNORECASE)))):
            items = line.split()
            jobId = items[0]
            j = index.jobSettings.get(jobId)
            if j is None:
                jobId_items = items[0].split('.')
                if jobId_items[0].upper() == 'TOTAL':
                    continue
                if len(jobId_items) == 1 or jobId_items[0].isdigit():
                    jobId = jobId_items[0]
                else:
                    jobId = jobId_items[1]
                j = index.jobSettings.get(jobId)
            if j is None:
                if warnNotBatch:
                    print('...Job ' + jobId + ' not in this batch, skipping')
                continue
            names += [j.names]
            job_names += [j.jobName]
            ids += [jobId]
            infos += [line]
    return ids, job_names, names, infos


def queue_job_names(batchPath=None, running=False, queued=True):
    lists = queue_job_details(batchPath, running, queued)[2]
    names = []
    for nameset in lists:
        names += nameset
    return names


# Functions to manipulate job templates ##################################################

def replace_placeholders(txt, vals):
    """
    Replaces placeholders ``vals`` (dict) in a template ``txt``.
    """
    txt = txt.replace('\r', '').format(**vals)
    return txt


def extract_value(txt, name):
    """
    Extracts value of variable ``name`` defined in a template ``txt``
    as ``##name: value``.
    """
    match = re.search('##' + name + ':(.*)##', txt)
    if match:
        return match.group(1).strip()
    return None


def get_defaulted(key_name, default=None, tp: Any = str, template=None, ext_env=None,
                  **kwargs) -> Any:
    val = kwargs.get(key_name)
    if val is None and template is not None:
        val = extract_value(template, 'DEFAULT_' + key_name)
    if val is None:
        val = os.environ.get(code_prefix + '_' + key_name, None)
    if val is None and ext_env:
        val = os.environ.get('ext_env', None)
    if val is None:
        val = default
    if val is None:
        return None
    return tp(val)



import os
from copy import deepcopy
from itertools import chain
from scipy.stats import multivariate_normal
import numpy as np
import pytest

from cobaya import run, load_samples
from cobaya.post import post, OutputOptions
from cobaya.typing import ParamsDict, InputDict
from cobaya.conventions import separator_files
from cobaya import mpi

pytestmark = pytest.mark.mpi

_post_ = separator_files + "post" + separator_files

mean = np.array([0, 0])
sigma = 0.5
cov = np.array([[sigma ** 2, 0], [0, sigma ** 2]])
sampled = {"mean": mean, "cov": cov}
target = {"mean": mean + np.array([sigma / 2, 0]), "cov": cov}


def allclose(a, b):
    return np.allclose(a.to_numpy(dtype=np.float64), b.to_numpy(dtype=np.float64))


def sampled_pdf(a, b):
    return multivariate_normal.logpdf([a, b], mean=sampled["mean"], cov=sampled["cov"])


def target_pdf(a, b, c=0):
    logp = multivariate_normal.logpdf([a, b], mean=target["mean"], cov=target["cov"])
    derived = {"cprime": c}
    return logp, derived


def target_pdf_prior(a, b):
    return target_pdf(a, b, c=0)[0]


_range = {"min": -2, "max": 2}
ref_pdf = {"dist": "norm", "loc": 0, "scale": 0.1}
info_params: ParamsDict = dict([
    ("a", {"prior": _range, "ref": ref_pdf, "proposal": sigma}),
    ("b", {"prior": _range, "ref": ref_pdf, "proposal": sigma}),
    ("a_plus_b", {"derived": lambda a, b: a + b})])

info_sampler = {"mcmc": {"Rminus1_stop": 0.25, "Rminus1_cl_stop": 0.5, "seed": 1}}
info_sampler_dummy = {"evaluate": {"N": 10}}


def _get_targets_cobaya(collections_in):
    # Notice that we do the importance reweighting, but do not change the logposterior
    # in the collection (not used for mean or cov anyway).
    loglikes = []
    for c in collections_in:
        a, b = c["a"].to_numpy(dtype=np.float64), c["b"].to_numpy(dtype=np.float64)
        loglikes.append(
            [target_pdf_prior(ai, bi) - sampled_pdf(ai, bi) for ai, bi in zip(a, b)]
        )
    # Subtract max over ALL collections: relative weights between collection would appear
    max_loglikes = max(list(chain(*loglikes)))
    importance_weights = [np.exp(np.array(like) - max_loglikes) for like in loglikes]
    collections_in[0].reweight(importance_weights, with_batch=collections_in[1:])
    # Merge before computing statistics
    for c in collections_in[1:]:
        collections_in[0]._append(c)  # noqa
    target_mean = collections_in[0].mean()
    target_cov = collections_in[0].cov()
    return target_mean, target_cov


def _get_targets_getdist(mcsamples_in):
    loglikes = []
    for a, b in zip(mcsamples_in['a'], mcsamples_in['b']):
        loglikes.append(-target_pdf_prior(a, b) + sampled_pdf(a, b))
    mcsamples_in.reweightAddingLogLikes(loglikes)
    target_mean = mcsamples_in.mean(["a", "b"])
    target_cov = mcsamples_in.getCovMat().matrix
    return target_mean, target_cov


@mpi.sync_errors
@pytest.mark.parametrize("temperature", (1, 2))
def test_post_prior(tmpdir, temperature):
    """
    Swaps prior "gaussian" for "target".

    It also tests loading vs passing samples to post, and different MCMC temperatures.
    """
    # Generate original chain
    info: InputDict = {
        "output": os.path.join(tmpdir, "gaussian"), "force": True,
        "params": info_params, "sampler": deepcopy(info_sampler),
        "likelihood": {"one": None}, "prior": {"gaussian": sampled_pdf}}
    info_post: InputDict = {
        "output": info["output"], "force": True,
        "post": {"suffix": "foo", 'skip': 0.1,
                 "remove": {"prior": {"gaussian": None}},
                 "add": {"prior": {"target": target_pdf_prior}}}}
    if list(info["sampler"].keys())[0] == "mcmc":
        if info["sampler"]["mcmc"] is None:
            info["sampler"]["mcmc"] = {}
        info["sampler"]["mcmc"]["temperature"] = temperature
    _, sampler = run(info)
    if mpi.is_main_process():
        collections_in = load_samples(info["output"], skip=0.1)
        target_mean_cobaya, target_cov_cobaya = mpi.share(
            _get_targets_cobaya(collections_in))
        mcsamples_in = load_samples(info["output"], skip=0.1, to_getdist=True)
        mcsamples_in.cool(temperature)
        target_mean_getdist, target_cov_getdist = mpi.share(
            _get_targets_getdist(mcsamples_in))
    else:
        target_mean_cobaya, target_cov_cobaya = mpi.share()
        target_mean_getdist, target_cov_getdist = mpi.share()
    for pass_chains in [False, True]:
        _, post_products = post(
            info_post, sample=sampler.samples() if pass_chains else None)
        # Load with GetDist and compare
        if mpi.is_main_process():
            mcsamples = load_samples(
                info_post["output"] + _post_ + info_post["post"]["suffix"],
                to_getdist=True,
            )
            new_mean = mcsamples.mean(["a", "b"])
            new_cov = mcsamples.getCovMat().matrix
            mpi.share((new_mean, new_cov))
        else:
            new_mean, new_cov = mpi.share()
        assert np.allclose(new_mean, target_mean_cobaya)
        assert np.allclose(new_cov, target_cov_cobaya)
        assert np.allclose(new_mean, target_mean_getdist)
        assert np.allclose(new_cov, target_cov_getdist)


def test_post_likelihood():
    """
    Swaps likelihood "gaussian" for "target".

    It also tests aggregated chi2's by removing and adding a likelihood to an existing
    type.
    """
    # Generate original chain
    orig_interval = OutputOptions.output_inteveral_s
    try:
        OutputOptions.output_inteveral_s = 0
        info_params_local = deepcopy(info_params)
        info_params_local["dummy"] = 0
        dummy_loglike_add = 0.1
        dummy_loglike_remove = 0.01
        info: InputDict = {
            "output": None, "force": True,
            "params": info_params_local, "sampler": info_sampler,
            "likelihood": {
                "gaussian": {"external": sampled_pdf, "type": "A"},
                "dummy": {"external": lambda dummy: 1, "type": "BB"},
                "dummy_remove": {"external": lambda dummy: dummy_loglike_add,
                                 "type": "BB"}}}
        info_out, sampler = run(info)
        mcsamples_in = sampler.samples(to_getdist=True)
        if not mpi.is_main_process():
            mcsamples_in = None

        info_out.update({
            "post": {"suffix": "foo",
                     "remove": {"likelihood": {
                         "gaussian": None, "dummy_remove": None}},
                     "add": {"likelihood": {
                         "target": {
                             "external": target_pdf, "type": "A",
                             "output_params": ["cprime"]},
                         "dummy_add": {
                             "external": lambda dummy: dummy_loglike_remove,
                             "type": "BB"}}}}})
        info_post_out, products_post = post(info_out, sampler.samples())
        mcsamples_post = products_post.samples(to_getdist=True)

        # Load with GetDist and compare
        if mcsamples_in:
            target_mean, target_cov = mpi.share(_get_targets_getdist(mcsamples_in))
            new_mean = mcsamples_post.mean(["a", "b"])
            new_cov = mcsamples_post.getCovMat().matrix
            mpi.share((new_mean, new_cov))
        else:
            target_mean, target_cov = mpi.share()
            new_mean, new_cov = mpi.share()
        assert np.allclose(new_mean, target_mean)
        assert np.allclose(new_cov, target_cov)
        assert allclose(products_post.samples(combined=True)["chi2__A"],
                        products_post.samples(combined=True)["chi2__target"])
        assert allclose(products_post.samples(combined=True)["chi2__BB"],
                        products_post.samples(combined=True)["chi2__dummy"] +
                        products_post.samples(combined=True)["chi2__dummy_add"])
    finally:
        OutputOptions.output_inteveral_s = orig_interval


def test_post_params(tmpdir):
    # Tests:
    # - added simple dynamical derived parameter "a+b"
    # - added dynamical derived parameter that depends on *new* chi2__target
    # - added new fixed input "c" + new derived-from-external-function "cprime"
    # Generate original chain
    info = {
        "params": info_params, "sampler": info_sampler_dummy,
        "likelihood": {"gaussian": sampled_pdf},
        "output": os.path.join(tmpdir, "post_params")}
    updated_info_gaussian, sampler_gaussian = run(info)
    info_post = {
        "post": {"suffix": "foo",
                 "remove": {"params": {"a_plus_b": None}},
                 "add": {
                     "likelihood": {
                         "target": {"external": target_pdf, "output_params": ["cprime"]}},
                     "params": {
                         "c": 1.234,
                         "a_minus_b": {"derived": "lambda a,b: a-b"},
                         "my_chi2__target": {
                             "derived": "lambda chi2__target: chi2__target"},
                         "cprime": None}}}}
    info_post.update(updated_info_gaussian)
    _, products = post(info_post, sampler_gaussian.products()["sample"])
    # Compare parameters
    assert allclose(products.samples(combined=True)["a"] -
                    products.samples(combined=True)["b"],
                    products.samples(combined=True)["a_minus_b"])
    assert np.allclose(
        products.samples(combined=True)["cprime"].to_numpy(dtype=np.float64),
        info_post["post"]["add"]["params"]["c"]
    )
    assert allclose(products.samples(combined=True)["my_chi2__target"],
                    products.samples(combined=True)["chi2__target"])
    # Same, but with loaded samples
    loaded_samples = load_samples(
        info_post["output"] + ".post." + info_post["post"]["suffix"], combined=True)
    assert allclose(products.samples(combined=True)["a"] -
                    products.samples(combined=True)["b"],
                    loaded_samples["a_minus_b"])
    assert np.allclose(loaded_samples["cprime"].to_numpy(dtype=np.float64),
                       info_post["post"]["add"]["params"]["c"])
    assert allclose(loaded_samples["my_chi2__target"],
                    products.samples(combined=True)["chi2__target"])



# Export the results to GetDist
gd_sample = sampler.products(to_getdist=True)["sample"]

# Analyze and plot
mean = gd_sample.getMeans()[:2]
covmat = gd_sample.getCovMat().matrix[:2, :2]
print("Mean:")
print(mean)
print("Covariance matrix:")
print(covmat)
# %matplotlib inline  # uncomment if running from the Jupyter notebook
import getdist.plots as gdplt

gdplot = gdplt.get_subplot_plotter()
gdplot.triangle_plot(gd_sample, ["a", "b"], filled=True)



from .pantheonplus import PantheonPlus


class Union3(PantheonPlus):
    """
    Likelihood for the Union3 & UNITY1.5 type Ia supernovae sample.

    Reference
    ---------
    https://arxiv.org/pdf/2311.12098.pdf
    """

    def configure(self):
        self.pre_vars = 0.  # diagonal component

    def _read_data_file(self, data_file):
        file_cols = ['zcmb', 'mb']
        self.cols = ['zcmb', 'mag']
        self._read_cols(data_file, file_cols)
        self.zhel = self.zcmb



"""
.. module:: cobaya.grid_tools.gridtables

:Synopsis: make summary pdf with tables for result in a grid (Cobaya version)
:Author: Antony Lewis (based on CosmoMC version of the same code)

"""
import os
import copy
from cobaya.tools import working_directory
from cobaya.grid_tools.batchjob_args import BatchArgs
from getdist import types, paramnames
from getdist.mcsamples import loadMCSamples


def grid_tables(args=None):

    opts = BatchArgs('Make pdf tables from latex generated from getdist outputs',
                     importance=True, converge=True)
    opts.parser.add_argument('latex_filename', help="name of latex/PDF file to produce")
    opts.parser.add_argument('--limit', type=int, default=2,
                             help="sigmas of quoted confidence intervals")
    opts.parser.add_argument('--all_limits', action='store_true')
    opts.parser.add_argument('--bestfitonly', action='store_true')
    opts.parser.add_argument('--nobestfit', action='store_true')
    opts.parser.add_argument('--no_delta_chisq', action='store_true')
    opts.parser.add_argument('--delta_chisq_paramtag', default=None,
                             help="parameter tag to give best-fit "
                                  "chi-squared differences")
    opts.parser.add_argument('--changes_from_datatag', default=None,
                             help="give fractional sigma shifts compared to "
                                  "a given data combination tag")
    opts.parser.add_argument('--changes_from_paramtag', default=None,
                             help="give fractional sigma shifts compared to a "
                                  "given parameter combination tag")
    opts.parser.add_argument('--changes_adding_data', nargs='+', default=None,
                             help="give fractional sigma shifts when adding given data")
    opts.parser.add_argument('--changes_replacing', nargs='+', default=None,
                             help='give sigma shifts for results with data x, y, z '
                                  'replacing data y, z.. with x')
    opts.parser.add_argument('--changes_only', action='store_true',
                             help='Only include results in the changes_replacing set')
    opts.parser.add_argument('--changes_data_ignore', nargs='+',
                             help='ignore these data tags when mapping to '
                                  'reference for comparison')
    opts.parser.add_argument('--systematic_average', action='store_true',
                             help='Combine two chains and quote results '
                                  'for the combination, e.g. as a crude way of '
                                  'including systematic errors between '
                                  'likelihood versions. '
                                  'Used with --changes_replacing or similar.')

    opts.parser.add_argument('--shift_sigma_indep', action='store_true',
                             help="fractional shifts are relative to the sigma for "
                                  "independent data (sigma^2=sigma1^2+sigma2^2")
    opts.parser.add_argument('--shift_sigma_subset', action='store_true',
                             help="fractional shifts are relative to the sigma "
                                  "for stricly subset data "
                                  "(sigma^2 = |sigma1^2-sigma2^2|, "
                                  "regularized to sigma/20)")

    opts.parser.add_argument('--paramNameFile',
                             help=".paramnames file for custom labels for parameters")

    opts.parser.add_argument('--paramList', default=None,
                             help=".paramnames file listing specific "
                                  "parameters to include (only).")
    opts.parser.add_argument('--blockEndParams', default=None,
                             help='a semi-colon separated list of parameters marking '
                                  'the end of distinct parameter blocks '
                                  '(e.g. physical vs nuisance parmeters, '
                                  'sampled vs derived)')
    opts.parser.add_argument('--columns', type=int, nargs=1, default=3)
    opts.parser.add_argument('--compare', nargs='+', default=None)

    opts.parser.add_argument('--titles', default=None)  # for compare plots
    opts.parser.add_argument('--forpaper', action='store_true')
    opts.parser.add_argument('--separate_tex', action='store_true')
    opts.parser.add_argument('--header_tex', default=None)
    opts.parser.add_argument('--height', default="13in")
    opts.parser.add_argument('--width', default="12in")

    (batch, args) = opts.parseForBatch(args)

    if args.blockEndParams is not None:
        args.blockEndParams = args.blockEndParams.split(';')

    if args.paramList is not None:
        args.paramList = paramnames.ParamNames(args.paramList)

    def tex_escape_text(string):
        return string.replace('_', '{\\textunderscore}')

    def get_table_lines(content, _referenceDataJobItem=None):
        if _referenceDataJobItem is not None:
            refResults = _referenceDataJobItem.result_marge
        else:
            refResults = None
        return types.ResultTable(args.columns, [content],
                                 blockEndParams=args.blockEndParams,
                                 paramList=args.paramList, limit=args.limit,
                                 refResults=refResults,
                                 shiftSigma_indep=args.shift_sigma_indep,
                                 shiftSigma_subset=args.shift_sigma_subset).lines

    def get_systematic_average_table_lines(jobItem1, jobItem2):
        # if you have two versions of the likelihood with the same data,
        # and don't know which is right,
        # this just crudely adds the samples with equal weight per likelihood
        samps1 = loadMCSamples(jobItem1.chainRoot, jobItem=jobItem1,
                               settings=batch.getdist_options)
        samps2 = loadMCSamples(jobItem2.chainRoot, jobItem=jobItem2,
                               settings=batch.getdist_options)
        samps = samps1.getCombinedSamplesWithSamples(samps2)
        return get_table_lines(samps.getMargeStats())

    def param_result_table(jobItem, deltaChisqJobItem=None, _referenceDataJobItem=None):
        if deltaChisqJobItem is not None and deltaChisqJobItem.name == jobItem.name:
            deltaChisqJobItem = None
        if _referenceDataJobItem is not None:
            if (args.changes_from_paramtag is None
                    and _referenceDataJobItem.normed_data == jobItem.normed_data
                    or args.changes_from_paramtag is not None
                    and _referenceDataJobItem.name == jobItem.name):
                _referenceDataJobItem = None

        table_lines = []
        caption = []
        jobItem.loadJobItemResults(paramNameFile=args.paramNameFile,
                                   bestfit=not args.nobestfit,
                                   bestfitonly=args.bestfitonly)
        bf = jobItem.result_bestfit
        if bf is not None:
            caption.append(
                ' Best-fit $\\chi^2_{\\rm eff} = ' + ('%.2f' % (bf.logLike * 2)) + '$')
            if deltaChisqJobItem is not None:
                bf_ref = deltaChisqJobItem.result_bestfit
                if bf_ref is not None:
                    caption.append(
                        '$\\Delta \\chi^2_{\\rm eff} = ' + (
                                '%.2f' % ((bf.logLike - bf_ref.logLike) * 2)) + '$')

        if args.bestfitonly:
            if bf is not None:
                table_lines += get_table_lines(bf)
        else:
            likeMarge = jobItem.result_likemarge
            if likeMarge is not None and likeMarge.meanLogLike is not None:
                caption.append('$\\bar{\\chi}^2_{\\rm eff} = ' + (
                        '%.2f' % (likeMarge.meanLogLike * 2)) + '$')
                if deltaChisqJobItem is not None:
                    likeMarge_ref = deltaChisqJobItem.result_likemarge
                    if likeMarge_ref is not None \
                            and likeMarge_ref.meanLogLike is not None:
                        delta = likeMarge.meanLogLike - likeMarge_ref.meanLogLike
                        caption.append('$\\Delta\\bar{\\chi}^2_{\\rm eff} = ' + (
                                '%.2f' % (delta * 2)) + '$')
            if jobItem.result_converge is not None and jobItem.result_converge.R_eigs:
                caption.append('$R-1 =' + jobItem.result_converge.worstR() + '$')
            if jobItem.result_marge is not None:
                if args.systematic_average:
                    table_lines += get_systematic_average_table_lines(
                        jobItem, _referenceDataJobItem)
                else:
                    table_lines += get_table_lines(jobItem.result_marge,
                                                   _referenceDataJobItem)

        table_lines.append('')
        if not args.forpaper:
            table_lines.append("; ".join(caption))
        if bf is not None and not args.forpaper:
            table_lines.append('')
            table_lines.append('$\\chi^2_{\\rm eff}$:')
            if deltaChisqJobItem is not None:
                compChiSq = deltaChisqJobItem.result_bestfit
            else:
                compChiSq = None
            for kind, vals in bf.sortedChiSquareds():
                table_lines.append(kind + ' - ')
                for val in vals:
                    line = '  ' + tex_escape_text(val.name) + ': ' + (
                            '%.2f' % val.chisq) + ' '
                    if compChiSq is not None:
                        comp = compChiSq.chiSquareForKindName(kind, val.name)
                        if comp is not None:
                            line += r'($\Delta$ ' + ('%.2f' % (val.chisq - comp)) + ') '
                    table_lines.append(line)
        return table_lines

    def compare_table(jobItems, titles=None):
        for job_i in jobItems:
            job_i.loadJobItemResults(paramNameFile=args.paramNameFile,
                                     bestfit=not args.nobestfit,
                                     bestfitonly=args.bestfitonly)
            print(job_i.name)
        if titles is None:
            titles = [job_i.datatag for job_i in jobItems if
                      job_i.result_marge is not None]
        else:
            titles = titles.split(';')
        return types.ResultTable(1,
                                 [job_i.result_marge for job_i in jobItems if
                                  job_i.result_marge is not None],
                                 limit=args.limit, titles=titles,
                                 blockEndParams=args.blockEndParams,
                                 paramList=args.paramList).lines

    if args.changes_replacing is not None:
        if args.data is not None:
            args.data += args.changes_replacing

    items = opts.sortedParamtagDict(chainExist=not args.bestfitonly)

    if args.all_limits:
        limits = [1, 2, 3]
    else:
        limits = [args.limit]

    if args.changes_from_paramtag is not None:
        if args.changes_from_datatag is not None:
            raise Exception(
                'You cannot have both changes_from_paramtag and changes_from_datatag')
        if args.delta_chisq_paramtag is not None \
                and args.delta_chisq_paramtag != args.changes_from_paramtag:
            raise Exception('when using changes_from_paramtag, '
                            'delta_chisq_paramtag is set equal to that')
        if args.no_delta_chisq:
            raise Exception('when using changes_from_paramtag cannot have no_delta_chisq')
        args.delta_chisq_paramtag = args.changes_from_paramtag

    def data_index(job_i):
        if args.changes_data_ignore:
            ignores = {}
            for ig in args.changes_data_ignore:
                ignores[ig] = ''
            return job_i.data_set.makeNormedDatatag(ignores)
        else:
            return job_i.normed_data

    baseJobItems = {}
    for paramtag, parambatch in items:
        isBase = len(parambatch[0].param_set) == 0
        for job_item in parambatch:
            if (args.delta_chisq_paramtag is None and
                    isBase and not args.no_delta_chisq or
                    args.delta_chisq_paramtag is not None
                    and job_item.paramtag == args.delta_chisq_paramtag):
                referenceJobItem = copy.deepcopy(job_item)
                referenceJobItem.loadJobItemResults(paramNameFile=args.paramNameFile)
                baseJobItems[job_item.normed_data] = referenceJobItem

    loc = os.path.split(args.latex_filename)[0]
    if loc:
        os.makedirs(loc, exist_ok=True)

    for limit in limits:
        args.limit = limit

        outfile = args.latex_filename
        if args.all_limits:
            outfile += '_limit' + str(limit)
        if outfile[-4:] != '.tex':
            outfile += '.tex'

        lines = []
        if not args.forpaper:
            lines.append('\\documentclass[10pt]{article}')
            lines.append('\\usepackage[pdftex]{hyperref}')
            lines.append('\\usepackage[paperheight=' + args.height +
                         ',paperwidth=' + args.width + ',margin=0.8in]{geometry}')
            lines.append('\\renewcommand{\\arraystretch}{1.5}')
            lines.append('\\begin{document}')
            if args.header_tex is not None:
                lines.append(open(args.header_tex, 'r').read())
            lines.append('\\tableofcontents')

        # set of baseline results, e.g. for Delta chi^2

        for paramtag, parambatch in items:
            isBase = len(parambatch[0].param_set) == 0
            if not args.forpaper:
                if isBase:
                    paramText = 'Baseline model'
                else:
                    paramText = tex_escape_text("+".join(parambatch[0].param_set))
                section = '\\newpage\\section{ ' + paramText + '}'
            else:
                section = ''
            if args.compare is not None:
                compares = opts.filterForDataCompare(parambatch, args.compare)
                if len(compares) == len(args.compare):
                    lines.append(section)
                    lines += compare_table(compares, args.titles)
                else:
                    print('no matches for compare: ' + paramtag)
            else:
                lines.append(section)
                theseItems = [jobItem for jobItem in parambatch
                              if (os.path.exists(jobItem.distPath) or args.bestfitonly)
                              and (args.converge == 0
                                   or jobItem.hasConvergeBetterThan(args.converge))]

                referenceDataJobItem = None
                if args.changes_from_datatag is not None:
                    for job_item in theseItems:
                        if (job_item.normed_data == args.changes_from_datatag
                                or job_item.datatag == args.changes_from_datatag):
                            referenceDataJobItem = copy.deepcopy(job_item)
                            referenceDataJobItem.loadJobItemResults(
                                paramNameFile=args.paramNameFile,
                                bestfit=args.bestfitonly)
                if args.changes_adding_data is not None:
                    baseJobItems = {}
                    refItems = []
                    for job_item in theseItems:
                        if job_item.data_set.hasName(args.changes_adding_data):
                            job_item.normed_without = "_".join(
                                sorted([x for x in job_item.data_set.names if
                                        x not in args.changes_adding_data]))
                            refItems.append(job_item.normed_without)
                        else:
                            job_item.normed_without = None
                    for job_item in theseItems:
                        if job_item.normed_data in refItems:
                            referenceJobItem = copy.deepcopy(job_item)
                            referenceJobItem.loadJobItemResults(
                                paramNameFile=args.paramNameFile,
                                bestfit=args.bestfitonly)
                            baseJobItems[job_item.normed_data] = referenceJobItem
                if args.changes_replacing is not None:
                    origCompare = [item for item in theseItems if
                                   args.changes_replacing[0] in item.data_set.names]
                    baseJobItems = {}
                    for job_item in origCompare:
                        referenceJobItem = copy.deepcopy(job_item)
                        referenceJobItem.loadJobItemResults(
                            paramNameFile=args.paramNameFile,
                            bestfit=args.bestfitonly)
                        baseJobItems[job_item.normed_data] = referenceJobItem

                for job_item in theseItems:
                    if args.changes_adding_data is not None:
                        if job_item.normed_without is not None:
                            referenceDataJobItem = baseJobItems.get(
                                job_item.normed_without, None)
                        else:
                            referenceDataJobItem = None
                        referenceJobItem = referenceDataJobItem
                        if args.changes_only and not referenceDataJobItem:
                            continue
                    elif args.changes_replacing is not None:
                        referenceDataJobItem = None
                        for replace in args.changes_replacing[1:]:
                            if replace in job_item.data_set.names:
                                referenceDataJobItem = baseJobItems.get(
                                    batch.normalizeDataTag(
                                        job_item.data_set.tagReplacing(
                                            replace, args.changes_replacing[0])), None)
                                break
                        referenceJobItem = referenceDataJobItem
                        if args.changes_only and not referenceDataJobItem:
                            continue
                    else:
                        referenceJobItem = baseJobItems.get(data_index(job_item), None)
                    if args.changes_from_paramtag is not None:
                        referenceDataJobItem = referenceJobItem
                    if args.systematic_average and referenceDataJobItem is None:
                        continue
                    if not args.forpaper:
                        if args.systematic_average:
                            lines.append('\\subsection{ ' + tex_escape_text(
                                job_item.name) + '/' + tex_escape_text(
                                referenceDataJobItem.name) + '}')
                        else:
                            lines.append(
                                '\\subsection{ ' + tex_escape_text(job_item.name) + '}')
                    try:
                        tableLines = param_result_table(job_item, referenceJobItem,
                                                        referenceDataJobItem)
                        if args.separate_tex:
                            types.TextFile(tableLines).write(job_item.distRoot + '.tex')
                        lines += tableLines
                    except Exception as e:
                        print('ERROR: ' + job_item.name)
                        print("Index Error:" + str(e))

        if not args.forpaper:
            lines.append('\\end{document}')

        (outdir, outname) = os.path.split(outfile)
        if outdir:
            os.makedirs(outdir, exist_ok=True)
        types.TextFile(lines).write(outfile)
        root = os.path.splitext(outname)[0]

        if not args.forpaper:
            print('Now converting to PDF...')
            delext = ['aux', 'log', 'out', 'toc']
            with working_directory(outdir):
                for _ in range(3):
                    # iterate three times to get table of contents page numbers right
                    os.system('pdflatex ' + outname)
                for ext in delext:
                    if os.path.exists(root + '.' + ext):
                        os.remove(root + '.' + ext)


def grid_param_compare(args=None):

    opts = BatchArgs('Compare parameter constraints over set of models, producing'
                     'latext column content for tables (but not full tables)')
    opts.parser.add_argument('--params', nargs='+',
                             help="list of parameters to compare in the output tables")
    opts.parser.add_argument('--single_extparams', nargs='+',
                             help="list of single parameter extensions to restrict to")

    opts.parser.add_argument('--compare', nargs='+', default=None,
                             help="list of data tags to compare")
    opts.parser.add_argument('--nobestfits', action='store_true')
    opts.parser.add_argument('--limit', type=int, default=2)
    opts.parser.add_argument('--latex_filename', default=None,
                             help="name of latex file to produce (otherwise stdout)")
    opts.parser.add_argument('--math_columns', action='store_true',
                             help="whether to include $$ in the output for each column")
    opts.parser.add_argument('--endline', default='\\cr')
    opts.parser.add_argument('--paramNameFile',
                             help="Optional parameter name file "
                                  "giving latex labels for each parameter")

    (batch, args) = opts.parseForBatch(args)
    formatter = types.TableFormatter()

    if args.paramNameFile:
        names = paramnames.ParamNames(args.paramNameFile)
        par_labels = dict((par, names.parWithName(par, True).label)
                          for par in names.names())
    else:
        par_labels = {}

    if args.compare:
        args.compare = [batch.normalizeDataTag(dat) for dat in args.compare]

    table = {}
    paramtag_for_param = {}
    for par in args.params:
        paramtag_for_param[par] = []

    for jobItem in opts.filteredBatchItems():
        if (args.compare is None or jobItem.matchesDatatag(args.compare)) and (
                not args.single_extparams or
                len(jobItem.param_set) == 1 and jobItem.hasParam(args.single_extparams)):
            jobItem.loadJobItemResults(paramNameFile=None, bestfit=not args.nobestfits,
                                       noconverge=True, silent=True)
            if jobItem.result_marge is not None:
                for par in args.params:
                    texValues = jobItem.result_marge.texValues(formatter, par,
                                                               limit=args.limit)
                    if texValues is not None:
                        if par not in par_labels:
                            par_labels[par] = \
                                jobItem.result_marge.parWithName(par, True).label
                        if jobItem.paramtag not in table:
                            table[jobItem.paramtag] = {}
                        data_table = table[jobItem.paramtag]
                        if jobItem.paramtag not in paramtag_for_param[par]:
                            paramtag_for_param[par].append(jobItem.paramtag)
                        if jobItem.normed_data not in data_table:
                            data_table[jobItem.normed_data] = {}
                        data_table[jobItem.normed_data][par] = texValues

    def make_math(txt):
        if args.math_columns:
            return '$' + txt + '$'
        else:
            return txt

    def text_as_column(txt, latex=True):
        if latex:
            return make_math(txt)
        else:
            return txt

    def sort_data(_batch, datatags):
        items = []
        for tag in datatags:
            items += [item for item in _batch if item == tag]
        return items

    lines = []

    for i, par in enumerate(args.params):
        for paramtag in paramtag_for_param[par]:
            dataTable = table[paramtag]
            print(paramtag, len(dataTable))
            if args.compare is not None:
                dataTable = sort_data(dataTable, args.compare)
            cols = [make_math(par_labels[par])]
            for datatag in dataTable:
                if args.latex_filename is not None:
                    for par2 in args.params:
                        if par2 in table[paramtag][datatag]:
                            res = table[paramtag][datatag][par2]
                            if len(res) > 1:
                                cols.append(text_as_column(res[1]))
                            cols.append(text_as_column(res[0], latex=res[0] != '---'))
                else:
                    print(' ', table[paramtag][datatag], datatag)
            lines.append(" & ".join(cols) + args.endline)

    if args.latex_filename is not None:
        if not args.latex_filename.endswith('.tex'):
            args.latex_filename += '.tex'
        (outdir, outname) = os.path.split(args.latex_filename)
        os.makedirs(os.path.dirname(outdir), exist_ok=True)
        types.TextFile(lines).write(args.latex_filename)


if __name__ == "__main__":
    grid_tables()



from .pantheonplus import PantheonPlus


class DESy5(PantheonPlus):
    """
    Likelihood for DES-Y5 type Ia supernovae sample.

    Reference
    ---------
    https://arxiv.org/abs/2401.02929
    """

    def configure(self):
        self.pre_vars = self.mag_err ** 2

    def _read_data_file(self, data_file):
        file_cols = ['zhd', 'zhel', 'mu', 'muerr_final']
        self.cols = ['zcmb', 'zhel', 'mag', 'mag_err']
        self._read_cols(data_file, file_cols, sep=',')



"""
.. module:: install

:Synopsis: Tools and script to install the external code and data packages needed by
           the Cobaya components to be used.
:Author: Jesus Torrado

"""
# Global
import os
import sys
import re
import subprocess
import traceback
import shutil
import tempfile
import logging
from itertools import chain
import importlib
from typing import List, Mapping, Union
import requests  # type: ignore
import tqdm  # type: ignore
from packaging import version

# Local
from cobaya.log import logger_setup, LoggedError, NoLogging, get_logger
from cobaya.component import get_component_class, ComponentNotFoundError
from cobaya.tools import create_banner, warn_deprecation, \
    write_packages_path_in_config_file, get_config_path, VersionCheckError, \
    resolve_packages_path, similar_internal_class_names, find_with_regexp
from cobaya.input import get_used_components
from cobaya.conventions import code_path, data_path, packages_path_arg, \
    packages_path_env, Extension, install_skip_env, packages_path_arg_posix, \
    packages_path_config_file, packages_path_input
from cobaya.mpi import set_mpi_disabled
from cobaya.typing import InputDict, InfoDict
from cobaya.yaml import yaml_load_file

_banner_symbol = "="
_banner_length = 80
_version_filename = "version.dat"


def do_package_install(component: str, package_install: Union[InfoDict, str],
                       full_code_path: str, logger):
    # attempt to install package if package_install download info is present
    # similar to InstallableLikelihood install_options (could be refactored)
    component_root = component.split('.')[0]
    directory = package_install.get("directory")
    min_version = package_install.get("min_version")
    if package_install == "pip":
        package_install = {"pip": None}
    elif not isinstance(package_install, Mapping):
        raise LoggedError(logger, "Invalid package_install: must be 'pip' or a "
                                  "dictionary with download_url or github_repository")
    package = package_install.get("pip")
    if package is None:
        package = component_root
    cwd = None
    if repo := package_install.get("github_repository"):
        logger.info('Downloading code from github (%s)', repo)
        directory = directory or repo.split("/")[-1]
        if not download_github_release(
                full_code_path, repo, package_install.get("github_release"),
                directory=directory, logger=logger):
            return False
        cwd = os.path.join(full_code_path, directory)
        package = '.'
    elif url := package_install.get("download_url"):
        logger.info('Downloading code from %s', url)
        cwd = os.path.join(full_code_path, directory or component_root)
        if not download_file(url, cwd, logger=logger):
            return False
        if not (paths := find_with_regexp('pyproject\\.toml', cwd, True)
                         or find_with_regexp('setup\\.py', cwd, True)):
            raise LoggedError(logger, "No setup.py or pyproject.toml "
                                      "found in %s for %s", cwd, component)
        cwd = os.path.dirname(paths[0])
        package = '.'
    elif "pip" not in package_install:
        raise LoggedError(logger, "Invalid package_install: must define pip, "
                                  "github_repository or download_url")
    if min_version is not None and package != ".":
        package += f">={min_version}"
    logger.info('pip installing %s', package)
    return not pip_install(package, upgrade=True, logger=logger, cwd=cwd)


def install(*infos, **kwargs):
    """
    Installs the external packages required by the components mentioned in ``infos``.

    ``infos`` can be input dictionaries, single component names, or names of yaml files

    :param force: force re-installation of apparently installed packages (default:
       ``False``).
    :param test: just check whether components are installed  (default: ``False``).
    :param upgrade: force upgrade of obsolete components (default: ``False``).
    :param skip: keywords of components that will be skipped during installation.
    :param skip_global: skip installation of already-available Python modules (default:
       ``False``).
    :param debug: produce verbose debug output  (default: ``False``).
    :param path: optional path where to install the packages
                 (defaults to any packages_path entry given in the info dictionaries).
    :param code: set to ``False`` to skip code packages (default: ``True``).
    :param data: set to ``False`` to skip data packages (default: ``True``).
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster) (default: ``False``).
    :param no_set_global: do not store the installation path for later runs (default:
       ``False``).

    :return: ``True`` if all components are installed without skips, ``False`` if
                  some components skipped, None (or exception raised) otherwise
    """
    debug = kwargs.get("debug", False)
    logger = kwargs.get("logger")
    if not logger:
        logger_setup(debug=debug)
        logger = get_logger("install")
    path = kwargs.get("path")
    infos = [yaml_load_file(info)
             if isinstance(info, str) and info.split('.')[-1] in Extension.yamls
             else info for info in infos]
    infos_not_single_names = [info for info in infos if isinstance(info, Mapping)]
    if not (path := path or resolve_packages_path(*infos_not_single_names)):
        raise LoggedError(
            logger, ("No 'path' argument given, and none could be found in input infos "
                     "(as %r), the %r env variable or the config file. "
                     "Maybe specify one via a command line argument '-%s [...]'?"),
            packages_path_input, packages_path_env, packages_path_arg[0])
    # General install path for all dependencies
    general_abspath = os.path.abspath(path)
    logger.info("Installing external packages at '%s'", general_abspath)
    # Set the installation path in the global config file
    if not kwargs.get("no_set_global") and not kwargs.get("test"):
        write_packages_path_in_config_file(general_abspath)
        logger.info(
            "The installation path has been written into the global config file: %s",
            os.path.join(get_config_path(), packages_path_config_file))
    kwargs_install = {"force": kwargs.get("force", False),
                      "no_progress_bars": kwargs.get("no_progress_bars"),
                      "show_error": False}
    for what in (code_path, data_path):
        kwargs_install[what] = kwargs.get(what, True)
        spath = os.path.join(general_abspath, what)
        if kwargs_install[what] and not os.path.exists(spath):
            try:
                os.makedirs(spath)
            except OSError:
                raise LoggedError(
                    logger, f"Could not create the desired installation folder '{spath}'")
    # To check e.g. for a version upgrade, it needs to reload the component class and
    # all relevant imported modules: the implementation of `is_installed` for each
    # class is expected to always reload external modules if passed `reload=True`
    # (should be False by default to avoid deleting objects unnecessarily).
    kwargs_is_installed = {"reload": True}
    unknown_components = []  # could not be identified
    failed_components = []  # general errors
    obsolete_components = []  # older or unknown version already installed
    skip_keywords_arg = set(kwargs.get("skip") or [])
    # NB: if passed with quotes as `--skip "a b"`, it's interpreted as a single key
    skip_keywords_arg = set(chain(*[word.split() for word in skip_keywords_arg]))
    skip_keywords_env = set(
        os.environ.get(install_skip_env, "").replace(",", " ").lower().split())
    skip_keywords = skip_keywords_arg.union(skip_keywords_env)
    skipped = False
    # Combine all requested components and install them
    # NB: components mentioned by name may be repeated with those given in dict infos.
    #     That's OK, because the install check will skip them in the 2nd pass
    used_components, components_infos = get_used_components(*infos, return_infos=True)
    for kind, components in used_components.items():
        for component in components:
            name_w_kind = (kind + ":" if kind else "") + component
            print()
            print(create_banner(name_w_kind,
                                symbol=_banner_symbol, length=_banner_length), end="")
            print()
            if _skip_helper(component.lower(), skip_keywords, skip_keywords_env, logger):
                skipped = True
                continue
            info = components_infos[component]
            if isinstance(info, str) or "external" in info:
                logger.info(
                    f"Component '{name_w_kind}' is a custom function. Nothing to do.")
                continue
            class_name = (info or {}).get("class")
            if class_name:
                logger.info(f"Class to be installed for this component: {class_name}")
            python_path = info.pop("python_path", None)
            package_install = info.get("package_install") or {}
            # If a class_name AND pip install were specified, make sure the class name is
            # used. Otherwise the pip package would be inferred from the alias.
            if (
                    package_install == "pip" or
                    ("pip" in package_install and package_install["pip"] is None)
            ):
                package_install["pip"] = class_name
            min_version = package_install.get('min_version')

            def _imported_class():
                return get_component_class(
                    component, kind=kind, component_path=python_path,
                    class_name=class_name, logger=logger, min_package_version=min_version,
                    allow_internal=not package_install)

            try:
                try:
                    if package_install and (
                            kwargs_install["force"] or kwargs.get("upgrade")):
                        raise ComponentNotFoundError(logger)
                    imported_class = _imported_class()
                except (ComponentNotFoundError, VersionCheckError):
                    if package_install:
                        if do_package_install(component, package_install,
                                              os.path.join(general_abspath, code_path),
                                              logger):
                            importlib.invalidate_caches()
                            imported_class = _imported_class()
                        else:
                            logger.error(
                                f"Package install failed for '{name_w_kind}'. Skipping.")
                            unknown_components += [name_w_kind]
                            continue
                    else:
                        raise
                except Exception:
                    raise
            except ComponentNotFoundError:
                logger.error(
                    f"Component '{name_w_kind}' could not be identified. Skipping.")
                unknown_components += [name_w_kind]
                continue
            except Exception:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                logger.error(f"An error occurred when loading '{name_w_kind}'. Skipping.")
                failed_components += [name_w_kind]
                continue
            else:
                if _skip_helper(imported_class.__name__.lower(), skip_keywords,
                                skip_keywords_env, logger):
                    skipped = True
                    continue
            # Update the name if the kind was unknown
            if not kind:
                name_w_kind = imported_class.get_kind() + ":" + component
            is_compatible = getattr(imported_class, "is_compatible", lambda: True)()
            if not is_compatible:
                logger.error(f"Skipping '{name_w_kind}' "
                             "because it is not compatible with your OS.")
                failed_components += [name_w_kind]
                continue
            logger.info("Checking if dependencies have already been installed...")
            is_installed = getattr(imported_class, "is_installed", None)
            if is_installed is None:
                logger.info(f"Component '{name_w_kind}' is a fully built-in component: "
                            "nothing to do.")
                continue
            this_component_install_path = general_abspath
            get_path = getattr(imported_class, "get_path", None)
            if get_path:
                this_component_install_path = get_path(this_component_install_path)
            # Check previous installations and their versions
            has_been_installed = False
            is_old_version_msg = None
            with NoLogging(None if debug else logging.ERROR):
                try:
                    if kwargs.get("skip_global"):
                        has_been_installed = is_installed(
                            path="global", **kwargs_install, **kwargs_is_installed)
                    if not has_been_installed:
                        has_been_installed = is_installed(
                            path=this_component_install_path, **kwargs_install,
                            **kwargs_is_installed)
                except VersionCheckError as excpt:
                    is_old_version_msg = str(excpt)
            if has_been_installed:  # no VersionCheckError was raised
                logger.info("External dependencies for this component already installed.")
                if kwargs.get("test"):
                    continue
                if kwargs_install["force"] and not kwargs.get("skip_global"):
                    logger.info("Forcing re-installation, as requested.")
                else:
                    logger.info("Doing nothing.")
                    continue
            elif is_old_version_msg:
                logger.info(f"Version check failed: {is_old_version_msg}")
                if kwargs.get("test"):
                    continue
                if not kwargs.get("upgrade") and not kwargs.get("force"):
                    logger.info("Skipping because '--upgrade' not requested.")
                    obsolete_components += [name_w_kind]
                    continue
            else:
                logger.info("Check found no existing installation")
                if not debug:
                    logger.info(
                        "(If you expected this to be already installed, re-run "
                        "`cobaya-install` with --debug to get more verbose output.)")
                if kwargs.get("test"):
                    # We are only testing whether it was installed, so consider it failed
                    failed_components += [name_w_kind]
                    continue
            # Do the install
            logger.info("Installing...")
            install_args = dict(kwargs_install, show_error=True)
            try:
                install_this = getattr(imported_class, "install", None)
                success = install_this(path=general_abspath, **install_args)
            except Exception:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                logger.error("An unknown error occurred. Delete the external packages "
                             "folder %r and try again. "
                             "Please, notify the developers if this error persists.",
                             general_abspath)
                success = False
            if success:
                logger.info("Successfully installed! Let's check it...")
            else:
                logger.error(
                    "Installation failed! Look at the error messages above. "
                    "Solve them and try again, or, if you are unable to solve them, "
                    "install the packages required by this component manually.")
                failed_components += [name_w_kind]
                continue
            # Test installation
            reloaded_class = get_component_class(
                component, kind=kind, component_path=info.pop("python_path", None),
                class_name=class_name, logger=logger)
            reloaded_is_installed = getattr(reloaded_class, "is_installed", None)
            with NoLogging(None if debug else logging.ERROR):
                try:
                    successfully_installed = reloaded_is_installed(
                        path=this_component_install_path, **kwargs_install,
                        **kwargs_is_installed)
                except Exception:
                    traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                    successfully_installed = False
            if not successfully_installed:
                logger.error("Installation apparently worked, "
                             "but the subsequent installation test failed! "
                             "This does not always mean that there was an actual error, "
                             "and is sometimes fixed simply by running the installer "
                             "again. If not, look closely at the error messages above, "
                             "or re-run with --debug for more verbose output. "
                             "If you are unable to fix the issues above, "
                             "try installing the packages required by this "
                             "component manually.")
                failed_components += [name_w_kind]
            else:
                logger.info("Installation check successful.")
    print()
    print(create_banner(" * Summary * ",
                        symbol=_banner_symbol, length=_banner_length), end="")
    print()
    bullet = "\n - "
    if unknown_components:
        suggestions_dict = {
            name: similar_internal_class_names(name) for name in unknown_components}
        suggestions_msg = \
            bullet + bullet.join(
                f"{name}: did you mean any of the following? {sugg} "
                "(mind capitalization!)" for name, sugg in suggestions_dict.items())
        raise LoggedError(
            logger, ("The following components could not be identified and were skipped:"
                     f"{suggestions_msg}"))
    if failed_components:
        raise LoggedError(
            logger, ("The installation (or installation test) of some component(s) has "
                     "failed: %s\nCheck output of the installer of each component above "
                     "for precise error info.\n"),
            bullet + bullet.join(failed_components))
    if obsolete_components:
        raise LoggedError(
            logger, ("The following packages are obsolete. Re-run with `--upgrade` option"
                     " (not upgrading by default to preserve possible user changes): %s"),
            bullet + bullet.join(obsolete_components))
    if not unknown_components and not failed_components and not obsolete_components:
        logger.info("All requested components' dependencies correctly installed at "
                    f"{general_abspath}")
        return not skipped


def _skip_helper(name, skip_keywords, skip_keywords_env, logger):
    try:
        this_skip_keyword = next(s for s in skip_keywords
                                 if s.lower() in name.lower())
        env_msg = (" in env var %r" % install_skip_env
                   if this_skip_keyword in skip_keywords_env else "")
        logger.info("Skipping %r as per skip keyword %r" + env_msg,
                    name, this_skip_keyword)
        return True
    except StopIteration:
        return False


def download_file(url, path, *, size=None, no_progress_bars=False, logger=None):
    """
    Downloads (and optionally decompresses) a file into a given path.

    :param url: url from which to download the file.
    :param path: path to where the file should be downloaded.
    :param size: size in bytes of the file to download; can be used to show percentages
       when the url headers do not show the actual size.
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster).
    :param logger: logger to use for reporting information; a new logger is created if not
       specified.
    :return: ``True`` if the download (and decompression, if requested) was successful,
       and ``False`` otherwise.
    """
    logger = logger or get_logger("install")
    with tempfile.TemporaryDirectory() as tmp_path:
        try:
            req = requests.get(url, allow_redirects=True, stream=True)
            # get hinted filename if available:
            try:
                filename = re.findall(
                    "filename=(.+)", req.headers['content-disposition'])[0]
                filename = filename.strip('"\'')
            except KeyError:
                filename = os.path.basename(url)
            filename_tmp_path = os.path.normpath(os.path.join(tmp_path, filename))
            size = size or int(req.headers.get('content-length', 0))
            # Adapted from https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
            if not no_progress_bars:
                bar = tqdm.tqdm(total=size, unit='iB', unit_scale=True, unit_divisor=1024)
            with open(filename_tmp_path, 'wb') as f:
                for data in req.iter_content(chunk_size=1024):
                    chunk_size = f.write(data)
                    if not no_progress_bars:
                        bar.update(chunk_size)
            if not no_progress_bars:
                bar.close()
            if os.path.getsize(filename_tmp_path) < 1024:  # 1kb
                with open(filename_tmp_path, "r") as f:
                    lines = f.readlines()
                    if lines[0].startswith("404") or "not found" in lines[0].lower():
                        raise ValueError("File not found (404)!")
            logger.info('Downloaded filename %s', filename)
        except Exception as excpt:
            logger.error(
                "Error downloading %s' to folder '%s': %s", url, tmp_path, excpt)
            return False
        logger.debug('Got: %s', filename)
        extension = os.path.splitext(filename)[-1][1:]
        if not os.path.exists(path):
            os.makedirs(path)
        try:
            if extension == "zip":
                from zipfile import ZipFile
                with ZipFile(filename_tmp_path, 'r') as zipObj:
                    zipObj.extractall(path)
            else:
                import tarfile
                if extension == "tgz":
                    extension = "gz"
                kwargs = {}
                if sys.version_info >= (12, 0):
                    kwargs['filter'] = 'data'
                with tarfile.open(filename_tmp_path, "r:" + extension) as tar:
                    tar.extractall(path, **kwargs)
            logger.debug('Decompressed: %s', filename)
            return True
        except Exception as excpt:
            logger.error("Error decompressing downloaded file! Corrupt file? [%s]", excpt)
            return False


def download_github_release(base_directory, repo_name, release_name=None, *, asset=None,
                            directory=None, no_progress_bars=False, logger=None):
    """
    Downloads a release (i.e. a tagged commit) or a release asset from a GitHub repo.

    :param base_directory: directory into which the release will be downloaded; will be
       created if it does not exist.
    :param repo_name: repository name as ``user/repo``.
    :param release_name: name or tag of the release.
                         If not set, uses latest default branch.
    :param asset: download just an asset (attached file) from a release.
    :param directory: name of the directory that will contain the asset or release, if
       different ``repo_name``.
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster).
    :param logger: logger to use for reporting information; a new logger is created if not
       specified.
    :return: ``True`` if the download was successful, and ``False`` otherwise.
    """
    logger = logger or get_logger("install")
    if "/" in repo_name:
        github_user = repo_name[:repo_name.find("/")]
        repo_name = repo_name[repo_name.find("/") + 1:]
    else:
        github_user = "CobayaSampler"
    base_url = r"https://github.com/" + github_user + "/" + repo_name
    download_directory = base_directory
    if not release_name:
        data = requests.get(
            "https://api.github.com/repos/" + github_user + "/" + repo_name).json()
        release_name = (data or {}).get('default_branch', 'main')
    if asset:
        url = (base_url + "/releases/download/" + release_name + "/" + asset)
        # Asset would get decompressed in base directory
        download_directory = os.path.join(download_directory, directory or repo_name)
    else:
        url = (base_url + "/archive/" + release_name + ".tar.gz")

    if not download_file(url, download_directory,
                         no_progress_bars=no_progress_bars, logger=logger):
        return False
    # In releases, not assets, remove version number from directory name
    # and rename if requested
    if not asset:
        w_version = next(d for d in os.listdir(base_directory)
                         if (d.startswith(repo_name) and len(d) != len(repo_name)))
        actual_download_directory = os.path.join(base_directory, w_version)
        download_directory = os.path.join(base_directory, directory or repo_name)
        if os.path.exists(download_directory):
            shutil.rmtree(download_directory)
        os.rename(actual_download_directory, download_directory)
    # Now save the version into a file to be checked later
    with open(os.path.join(download_directory, _version_filename), "w") as f:
        f.write(release_name)
    logger.info((f"{asset} from " if asset else "") +
                "%s %s downloaded and decompressed correctly.", repo_name, release_name)
    return True


def pip_install(packages, upgrade=False, logger=None, options=(), **kwargs):
    """
    Takes package name or list of them.

    Returns exit status.
    """
    if hasattr(packages, "split"):
        packages = [packages]
    cmd = [sys.executable, '-m', 'pip', 'install']
    if upgrade:
        cmd += ['--upgrade']
    cmd += list(options)
    # Assume that if the user has installed Cobaya on the system-wide Python,
    # then it is OK to overwrite system packages:
    kwargs["env"] = dict(kwargs.get("env", os.environ), PIP_BREAK_SYSTEM_PACKAGES="1")
    res = subprocess.call(cmd + packages, **kwargs)
    if res:
        msg = f"pip: error installing packages '{packages}'"
        logger.error(msg) if logger else print(msg, file=sys.stderr)
    return res


def check_gcc_version(min_version="6.4", error_returns=None):
    """
    Checks the version of the ``gcc`` compiler installed.

    If the installed version is higher than ``min_version``, return ``True``,
    and ``False`` otherwise.

    If an error is produced, returns ``error_returns``.
    """
    try:
        ver = subprocess.check_output(
            "gcc -dumpversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    except subprocess.CalledProcessError:
        return error_returns
    # Change in gcc >= 7: -dumpversion only dumps major version
    if "." not in ver:
        ver = subprocess.check_output(
            "gcc -dumpfullversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    return version.parse(str(min_version)) <= version.parse(ver)


# Command-line script ####################################################################

def install_script(args=None):
    """Command line script for the installer."""
    set_mpi_disabled()
    warn_deprecation()
    # Parse arguments
    import argparse
    parser = argparse.ArgumentParser(
        prog="cobaya install",
        description="Cobaya's installation tool for external packages.")
    parser.add_argument("files_or_components", action="store", nargs="+",
                        metavar="input_file.yaml|component_name",
                        help="One or more input files or component names "
                             "(or simply 'cosmo' to install all the requisites for basic"
                             " cosmological runs)")
    parser.add_argument("-" + packages_path_arg[0], "--" + packages_path_arg_posix,
                        action="store", required=False,
                        metavar="/packages/path", default=None,
                        help="Desired path where to install external packages. "
                             "Optional if one has been set globally or as an env variable"
                             " (run with '--show_%s' to check)." %
                             packages_path_arg_posix)
    output_show_packages_path = resolve_packages_path()
    if output_show_packages_path and os.environ.get(packages_path_env):
        output_show_packages_path += " (from env variable %r)" % packages_path_env
    elif output_show_packages_path:
        output_show_packages_path += " (from config file)"
    else:
        output_show_packages_path = "(Not currently set.)"
    parser.add_argument("--show-" + packages_path_arg_posix, action="version",
                        version=output_show_packages_path,
                        help="Prints default external packages installation folder "
                             "and exits.")
    parser.add_argument("-" + "f", "--" + "force", action="store_true", default=False,
                        help="Force re-installation of apparently installed packages.")
    parser.add_argument("--%s" % "test", action="store_true", default=False,
                        help="Just check whether components are installed.")
    parser.add_argument("--upgrade", action="store_true", default=False,
                        help="Force upgrade of obsolete components.")
    parser.add_argument("--skip", action="store", nargs="*",
                        metavar="keyword", help=("Keywords of components that will be "
                                                 "skipped during installation."))
    parser.add_argument("--skip-global", action="store_true", default=False,
                        help="Skip installation of already-available Python modules.")
    parser.add_argument("-" + "d", "--" + "debug", action="store_true",
                        help="Produce verbose debug output.")
    group_just = parser.add_mutually_exclusive_group(required=False)
    group_just.add_argument("-C", "--just-code", action="store_false", default=True,
                            help="Install code of the components.", dest=data_path)
    group_just.add_argument("-D", "--just-data", action="store_false", default=True,
                            help="Install data of the components.", dest=code_path)
    parser.add_argument("--no-progress-bars", action="store_true", default=False,
                        help=("No progress bars shown; use when output is saved into a "
                              "text file (e.g. when running on a cluster)."))
    parser.add_argument("--no-set-global", action="store_true", default=False,
                        help="Do not store the installation path for later runs.")
    arguments = parser.parse_args(args)
    # Configure the logger ASAP
    logger_setup(arguments.debug)
    logger = get_logger("install")
    # Gather requests
    infos: List[Union[InputDict, str]] = []
    for f in arguments.files_or_components:
        if f.lower() == "cosmo":
            logger.info("Installing basic cosmological packages.")
            from cobaya.cosmo_input import install_basic
            infos += [install_basic]
        elif f.lower() == "cosmo-tests":
            logger.info("Installing *tested* cosmological packages.")
            from cobaya.cosmo_input import install_tests
            infos += [install_tests]
        elif os.path.splitext(f)[1].lower() in Extension.yamls:
            from cobaya.input import load_input
            infos += [load_input(f)]
        else:  # a single component name, no kind specified
            infos += [f]
    # Launch installer
    install(*infos, path=getattr(arguments, packages_path_arg), logger=logger,
            **{arg: getattr(arguments, arg)
               for arg in ["force", code_path, data_path, "no_progress_bars", "test",
                           "no_set_global", "skip", "skip_global", "debug", "upgrade"]})


if __name__ == '__main__':
    install_script()



info_txt = r"""
likelihood:
  planck_2018_lowl.TT:
  planck_2018_lowl.EE:
  planck_NPIPE_highl_CamSpec.TTTEEE: null
  planck_2018_lensing.native: null
theory:
  classy:
    extra_args: {N_ur: 2.0328, N_ncdm: 1}
params:
  logA:
    prior: {min: 2, max: 4}
    ref: {dist: norm, loc: 3.05, scale: 0.001}
    proposal: 0.001
    latex: \log(10^{10} A_\mathrm{s})
    drop: true
  A_s: {value: 'lambda logA: 1e-10*np.exp(logA)', latex: 'A_\mathrm{s}'}
  n_s:
    prior: {min: 0.8, max: 1.2}
    ref: {dist: norm, loc: 0.96, scale: 0.004}
    proposal: 0.002
    latex: n_\mathrm{s}
  H0:
    prior: {min: 40, max: 100}
    ref: {dist: norm, loc: 70, scale: 2}
    proposal: 2
    latex: H_0
  omega_b:
    prior: {min: 0.005, max: 0.1}
    ref: {dist: norm, loc: 0.0221, scale: 0.0001}
    proposal: 0.0001
    latex: \Omega_\mathrm{b} h^2
  omega_cdm:
    prior: {min: 0.001, max: 0.99}
    ref: {dist: norm, loc: 0.12, scale: 0.001}
    proposal: 0.0005
    latex: \Omega_\mathrm{c} h^2
  m_ncdm: {renames: mnu, value: 0.06}
  Omega_Lambda: {latex: \Omega_\Lambda}
  YHe: {latex: 'Y_\mathrm{P}'}
  tau_reio:
    prior: {min: 0.01, max: 0.8}
    ref: {dist: norm, loc: 0.06, scale: 0.01}
    proposal: 0.005
    latex: \tau_\mathrm{reio}
"""

from cobaya.yaml import yaml_load

info = yaml_load(info_txt)

# Add your external packages installation folder
info['packages_path'] = '/path/to/packages'



from cobaya.model import get_model

model = get_model("sample_r_theta.yaml")

# Get one random point (select list element [0]).
# When sampling a random point, we need to ignore the Jacobian and
# the gaussian band, since Cobaya doesn't know how to sample from them
random_point = model.prior.sample(ignore_external=True)[0]

# Our random point is now an array. Turn it into a dictionary:
sampled_params_names = model.parameterization.sampled_params()
random_point_dict = dict(zip(sampled_params_names, random_point))
print("")
print("Our random point is:", random_point_dict)

# Let's print some probabilities for our random point
print("The log-priors are:", model.logpriors(random_point_dict, as_dict=True))
print("The log-likelihoods and derived parameters are:",
      model.loglikes(random_point_dict, as_dict=True))
print("The log-posterior is:", model.logpost(random_point_dict))

print("")
print("You can also get all that information at once!")
posterior_dict = model.logposterior(random_point_dict, as_dict=True)
for k, v in posterior_dict.items():
    print(k, ":", v)

import numpy as np
import matplotlib.pyplot as plt
rs = np.linspace(0.75, 1.25, 200)
loglikes = [model.loglike({"r": r, "theta": np.pi / 4}, return_derived=False)
            for r in rs]
plt.figure()
plt.plot(rs, np.exp(loglikes))
plt.savefig('model_slice.png')

# Optional: define an output driver
from cobaya.output import get_output
out = get_output(prefix="chains/my_model", resume=False, force=True)

# Initialise and run the sampler (low number of samples, as an example)
info_sampler = {"mcmc": {"max_samples": 100}}
from cobaya.sampler import get_sampler
mcmc = get_sampler(info_sampler, model=model, output=out)
mcmc.run()

# Print results
print(mcmc.products()["sample"])



# Request H(z)
import numpy as np

redshifts = np.linspace(0, 2.5, 40)
model.add_requirements({"Hubble": {"z": redshifts}})

omega_cdm = [0.10, 0.11, 0.12, 0.13, 0.14]

f, (ax_cl, ax_h) = plt.subplots(1, 2, figsize=(14, 6))
for o in omega_cdm:
    point["omega_cdm"] = o
    model.logposterior(point)  # to force computation of theory
    Cls = model.provider.get_Cl(ell_factor=True)
    ax_cl.plot(Cls["ell"][2:], Cls["tt"][2:], label=r"$\Omega_\mathrm{CDM}h^2=%g$" % o)
    H = model.provider.get_Hubble(redshifts)
    ax_h.plot(redshifts, H / (1 + redshifts), label=r"$\Omega_\mathrm{CDM}h^2=%g$" % o)
ax_cl.set_ylabel(r"$\ell(\ell+1)/(2\pi)\,C_\ell\;(\mu \mathrm{K}^2)$")
ax_cl.set_xlabel(r"$\ell$")
ax_cl.legend()
ax_h.set_ylabel(r"$H/(1+z)\;(\mathrm{km}/\mathrm{s}/\mathrm{Mpc}^{-1})$")
ax_h.set_xlabel(r"$z$")
ax_h.legend()
plt.savefig("omegacdm.png")
# plt.show()



info = {
    "likelihood": {
        "gaussian_mixture": {
            "means": [0.2, 0],
            "covs": [[0.1, 0.05],
                     [0.05, 0.2]],
            "derived": True}},
    "params": dict([
        ("a", {
            "prior": {"min": -0.5, "max": 3},
            "latex": r"\alpha"}),
        ("b", {
            "prior": {"dist": "norm", "loc": 0, "scale": 1},
            "ref": 0,
            "proposal": 0.5,
            "latex": r"\beta"}),
        ("derived_a", {
            "latex": r"\alpha^\prime"}),
        ("derived_b", {
            "latex": r"\beta^\prime"})]),
    "sampler": {
        "mcmc": None}}



# Minimization of a random Gaussian likelihood using the minimize sampler.

import os
import warnings
import numpy as np
import pytest

from cobaya import mpi, run, InputDict, Likelihood
from cobaya.component import ComponentNotInstalledError
from cobaya.samplers.minimize import valid_methods

pytestmark = pytest.mark.mpi

mean = np.array([0.30245268, 0.61884443, 0.5])
cov = np.array(
    [[0.00796336, -0.0014805, -0.00479433], [-0.0014805, 0.00561415, 0.00434189],
     [-0.00479433, 0.00434189, 0.03208593]])
_inv_cov = np.linalg.inv(cov)


class NoisyCovLike(Likelihood):
    params = {'a': [0, 1, 0.5, 0.3, 0.08], 'b': [0, 1, 0.5, 0.3, 0.08],
              'c': [0, 1, 0.5, 0.3, 0.08]}
    noise = 0

    def logp(self, **params_values):
        x = np.array([params_values['a'], params_values['b'], params_values['c']]) - mean
        return -_inv_cov.dot(x).dot(x) / 2 + self.noise * np.random.random_sample()


@mpi.sync_errors
def test_minimize_gaussian(tmpdir, skip_not_installed):
    maxloglik = 0
    for method in reversed(valid_methods):
        NoisyCovLike.noise = 0.005 if method == 'bobyqa' else 0
        info: InputDict = {'likelihood': {'like': NoisyCovLike},
                           "sampler": {"minimize": {"ignore_prior": True,
                                                    "method": method}}}
        try:
            products = run(info)[1].products()
        except ComponentNotInstalledError:
            if skip_not_installed:
                warnings.warn(
                    f"Could not test minimization method '{method}': not installed."
                )
                continue
            raise
        error = abs(maxloglik - -products["minimum"]["minuslogpost"])
        assert error < 0.01
        info['output'] = os.path.join(tmpdir, 'testmin')
        products = run(info, force=True)[1].products()
        if mpi.is_main_process():
            from getdist.types import BestFit
            res = BestFit(info['output'] + '.bestfit').getParamDict()
            assert np.isclose(res["loglike"], products["minimum"]["minuslogpost"])
            for p, v in list(res.items())[:-2]:
                assert np.isclose(products["minimum"][p], v)
            assert isinstance(products["full_set_of_mins"], dict)


@mpi.sync_errors
def test_minimize_single_point(tmpdir, skip_not_installed):
    if mpi.get_mpi_size() > 1:
        return
    for method in reversed(valid_methods):
        NoisyCovLike.noise = 0.005 if method == 'bobyqa' else 0
        info: InputDict = {'likelihood': {'like': NoisyCovLike},
                           "sampler": {"minimize": {"ignore_prior": True,
                                                    "method": method,
                                                    "best_of": 1}},
                           'output': os.path.join(tmpdir, 'testmin')}
        try:
            products = run(info, force=True)[1].products()
        except ComponentNotInstalledError:
            if skip_not_installed:
                warnings.warn(
                    f"Could not test minimization method '{method}': not installed."
                )
                continue
            raise
        if mpi.is_main_process():
            assert products["full_set_of_mins"] is None


@mpi.sync_errors
def test_run_minimize(tmpdir):
    NoisyCovLike.noise = 0
    info: InputDict = {'likelihood': {'like': NoisyCovLike},
                       "sampler": {"mcmc": {"Rminus1_stop": 0.5, 'Rminus1_cl_stop': 0.4,
                                            'seed': 2}},
                       "output": os.path.join(tmpdir, 'testchain')}
    # No need to use skip_not_installed: uses default method, which should be requisite
    run(info, force=True)
    min_info: InputDict = dict(info, sampler={'minimize': None})
    output_info, sampler = run(min_info, force=True)
    assert (abs(sampler.products()["minimum"]["b"] - mean[1]) < 0.01)



"""
Allows calling `cobaya-[command]` as `python -m cobaya [command]`.

`run` is optional: one can pass directly an input file as
`python -m cobaya input.yaml`.
"""
import sys
from importlib import import_module, metadata


def run_command():
    commands = {}
    prefix = "cobaya-"
    console_scripts = (
        metadata.entry_points().select(group="console_scripts")
        if sys.version_info > (3, 9)
        else metadata.entry_points()["console_scripts"]
    )
    for script in console_scripts:
        if script.name.startswith(prefix):
            commands[script.name] = script.value
    commands_trimmed = [c[len(prefix):] for c in commands]
    help_msg = (
        "Add a one of the following commands and its arguments "
        f"(`<command> -h` for help): {commands_trimmed}"
    )
    try:
        command_or_input = sys.argv[1].lower()
    except IndexError:  # no command
        print(help_msg)
    else:
        command = commands.get("cobaya-" + command_or_input)
        if command is not None:
            module, func = command.split(":")
            sys.argv.pop(1)
            assert func is not None
            getattr(import_module(module), func)()
        else:
            if command_or_input in ["-h", "--help"]:
                print(help_msg)
            else:
                # no command --> assume run with input file as 1st arg (don't pop!)
                getattr(import_module("cobaya.run"), "run_script")()


if __name__ == "__main__":
    run_command()



"""
.. module:: cobaya.grid_tools.gridconfig

:Synopsis: Grid creator (Cobaya version)
:Author: Antony Lewis and Jesus Torrado
         (based on Antony Lewis' CosmoMC version of the same code)

"""

# Global
import os
import argparse
import importlib.util
from itertools import chain
from getdist.inifile import IniFile
from getdist.paramnames import makeList as make_list

# Local
from cobaya.yaml import yaml_load_file, yaml_dump_file
from cobaya.conventions import Extension, packages_path_input
from cobaya.input import get_used_components, merge_info, update_info
from cobaya.install import install as install_reqs
from cobaya.tools import sort_cosmetic, warn_deprecation, resolve_packages_path
from cobaya.grid_tools import batchjob
from cobaya.cosmo_input import create_input, get_best_covmat_ext, \
    get_covmat_package_folders
from cobaya.parameterization import is_sampled_param


def get_args(vals=None):
    parser = argparse.ArgumentParser(
        prog="cobaya-grid-create",
        description='Initialize grid using settings file')
    parser.add_argument('batchPath', help=(
        'root directory containing/to contain the grid '
        '(e.g. grid_folder where output directories are created '
        'at grid_folder/base/base_xx)'))
    parser.add_argument('settingName', nargs='?', help=(
        'python setting file for making or updating a grid, a py filename or '
        'full name of a python module'))
    parser.add_argument('--read-only', action='store_true', help=(
        'option to configure an already-run existing grid'))
    # Arguments related to installation of requisites
    parser.add_argument('--install-reqs-at', help=(
        'install required code and data for the grid in the given folder.'))
    parser.add_argument("--install-reqs-force", action="store_true", default=False,
                        help="Force re-installation of apparently installed packages.")
    parser.add_argument("--show-covmats", action="store_true",
                        help="Show which covmat is assigned to each chain.")

    return parser.parse_args(vals)


def path_is_grid(batchPath):
    return os.path.exists(batchjob.grid_cache_file(batchPath)) or os.path.exists(
        os.path.join(batchPath, 'config', 'config.ini'))


def grid_create(args=None):
    warn_deprecation()
    args = get_args(args)
    args.interactive = True
    makeGrid(**args.__dict__)


def import_from_path(full_path):
    # Create a module spec from the full path
    spec = importlib.util.spec_from_file_location(
        os.path.splitext(os.path.basename(full_path))[0], full_path)
    # Create a module from the spec
    module = importlib.util.module_from_spec(spec)
    # Execute the module to populate it
    spec.loader.exec_module(module)
    return module


def post_merge_info(*infos):
    # merge contents of add and remove, or if neither, assume should be options for "add"
    adds = []
    removes = []
    result = {}
    for info in infos:
        inf = info.copy()
        if "add" in info:
            adds.append(inf.pop("add"))
        if "remove" in info:
            removes.append(inf.pop("remove"))
        if len(inf) == len(info):
            adds.append(inf)
        else:
            result.update(inf)
    if adds:
        result["add"] = merge_info(*adds)
    if removes:
        result["remove"] = merge_info(*removes)
    return result


def set_minimize(info, minimize_info=None):
    result = dict(info, sampler={'minimize': minimize_info}, force=True)
    result.pop("resume", None)
    return result


# noinspection PyUnboundLocalVariable
def makeGrid(batchPath, settingName=None, settings=None, read_only=False,
             interactive=False, install_reqs_at=None, install_reqs_force=None,
             show_covmats=False):
    print("Generating grid...")
    batchPath = os.path.abspath(batchPath) + os.sep
    if not settings:
        if not settingName:
            if not path_is_grid(batchPath):
                raise Exception('Need to give name of setting file if batchPath/config '
                                'does not exist')
            read_only = True
            settingName = IniFile(os.path.join(batchPath + 'config',
                                               'config.ini')).params['setting_file']
            settingName = os.path.join(batchPath + 'config', settingName)
            if settingName.endswith('.py'):
                settings = import_from_path(settingName)
            else:
                settings = yaml_load_file(settingName)
        elif os.path.splitext(settingName)[-1].lower() in Extension.yamls:
            settings = yaml_load_file(settingName)
        elif settingName.endswith('.py'):
            settings = import_from_path(settingName)
        else:
            settings = __import__(settingName, fromlist=['dummy'])
            settingName = settings.__file__
    batch = batchjob.BatchJob(batchPath)
    batch.make_items(settings, messages=not read_only)
    if read_only:
        for job_item in batch.jobItems.copy():
            if not job_item.chainExists():
                batch.jobItems.remove(job_item)
        batch.save()
        print('OK, configured grid with %u existing chains' % (len(batch.jobItems)))
        return batch
    else:
        batch.make_directories(settingName or settings.__file__)
        batch.save()
    infos = {}
    components_used = {}

    from_yaml = isinstance(settings, dict)
    dic = settings if from_yaml else settings.__dict__
    yaml_dir = dic.get("yaml_dir") or ""
    if 'start_at_bestfit' in dic:
        raise ValueError("start_at_bestfit not yet implemented")

    def dicts_or_load(_infos):
        if not _infos or isinstance(_infos, dict):
            return [_infos or {}]
        return [(yaml_load_file(os.path.join(yaml_dir, _info)) if
                 isinstance(_info, str) else _info) for _info in _infos]

    def dict_option(_name):
        s = dic.get(_name) or {}
        if isinstance(s, str):
            return yaml_load_file(os.path.join(yaml_dir, s))
        return s

    defaults = merge_info(*dicts_or_load(dic.get('defaults')))
    importance_defaults = merge_info(*dicts_or_load(dic.get('importance_defaults')))
    minimize_defaults = merge_info(*dicts_or_load(dic.get('minimize_defaults')))
    params = dict_option('params')
    param_extra = dict_option('param_extra_opts')

    for job_item in batch.items(wantSubItems=False):
        # Model info
        job_item.makeChainPath()
        if (model_info := job_item.model_info) is None:
            model_info = {'params': {}}
            for par in job_item.param_set:
                if par not in params:
                    raise ValueError("params[%s] must be defined." % par)
                model_info['params'][par] = params[par]
            extra = dict(param_extra, **job_item.param_extra_opts)
            if opts := extra.get(job_item.paramtag):
                extra_infos = [opts]
            else:
                extra_infos = [extra[par] for par in job_item.param_set if par in extra]
            model_info = merge_info(job_item.defaults, model_info,
                                    *extra_infos)

        data_infos = dicts_or_load(job_item.data_set.infos)
        combined_info = merge_info(defaults, model_info, *data_infos)
        if "preset" in combined_info:
            preset = combined_info.pop("preset")
            combined_info = merge_info(create_input(**preset), combined_info)
        combined_info["output"] = job_item.chainRoot
        # Requisites
        components_used = get_used_components(components_used, combined_info)
        if install_reqs_at:
            combined_info[packages_path_input] = os.path.abspath(install_reqs_at)
        # Save the info (we will write it after installation:
        # we need to install to add auto covmats
        if job_item.paramtag not in infos:
            infos[job_item.paramtag] = {}
        infos[job_item.paramtag][job_item.data_set.tag] = combined_info
    # Installing requisites
    if install_reqs_at:
        print("Installing required code and data for the grid.")
        from cobaya.log import logger_setup
        logger_setup()
        install_reqs(components_used, path=install_reqs_at, force=install_reqs_force)
    print("Adding covmats (if necessary) and writing input files")
    cov_dir = dic.get("cov_dir")  # None means use the default from mcmc settings
    def_packages = cov_dir or install_reqs_at or resolve_packages_path()
    for job_item in batch.items(wantSubItems=False):
        info = infos[job_item.paramtag][job_item.data_set.tag]
        # Covariance matrices
        # We try to find them now, instead of at run time, to check if correctly selected
        try:
            sampler = list(info["sampler"])[0]
        except KeyError:
            raise ValueError("No sampler has been chosen: %s" % job_item.name)
        if sampler == "mcmc" and (cov_dir or cov_dir is None and
                                  info["sampler"][sampler].get("covmat") == "auto"):
            if not (cov_dirs := make_list(cov_dir or [])):
                if not (packages_path := install_reqs_at or info.get(packages_path_input)
                                         or def_packages):
                    raise ValueError(
                        "Cannot assign automatic covariance matrices because no "
                        "external packages path has been defined.")
                cov_dirs = get_covmat_package_folders(os.path.abspath(packages_path))
            # Need updated info for covmats: includes renames
            updated_info = update_info(info)
            # Ideally, we use slow+sampled parameters to look for the covariance matrix
            # but since for that we'd need to initialise a model, we approximate that set
            # as theory+sampled
            like_params = set(chain(*[
                list(like.get("params") or [])
                for like in updated_info["likelihood"].values()]))
            params_info = {p: v for p, v in updated_info["params"].items()
                           if is_sampled_param(v) and p not in like_params}

            best_covmat = get_best_covmat_ext(cov_dirs, params_info,
                                              updated_info["likelihood"],
                                              job_item=job_item,
                                              cov_map=dic.get("cov_map") or {})
            info["sampler"][sampler]["covmat"] = os.path.join(
                best_covmat["folder"], best_covmat["name"]) if best_covmat else None
            if show_covmats:
                print(job_item.name, '->', (best_covmat or {}).get("name"))

        # Write the info for this job
        # Allow overwrite since often will want to regenerate grid with tweaks
        info = sort_cosmetic(info)
        yaml_dump_file(job_item.yaml_file(), info, error_if_exists=False)

        # Minimize
        info = set_minimize(info, minimize_defaults)
        yaml_dump_file(job_item.yaml_file('_minimize'), info, error_if_exists=False)

        # Importance sampling
        for imp in job_item.importanceJobs():
            if getattr(imp, 'importanceFilter', None):
                continue
            if batch.hasName(imp.name.replace('.post.', '_')):
                raise Exception('importance sampling something you already have?')
            info_post = {"output": job_item.chainRoot,
                         "post": post_merge_info(importance_defaults,
                                                 *dicts_or_load(imp.importanceSettings)),
                         "force": True}
            info_post["post"]["suffix"] = imp.importanceTag
            yaml_dump_file(imp.yaml_file(), info_post, error_if_exists=False)
            if getattr(imp, 'want_minimize', True):
                info = set_minimize(dict(info, **info_post), minimize_defaults)
                yaml_dump_file(imp.yaml_file('_minimize'), info, error_if_exists=False)

    if not interactive:
        return batch
    print('Done... to run do: cobaya-grid-run %s' % batchPath)
    print('....... for best fits: cobaya-grid-run %s --minimize' % batchPath)
    print('For importance sampled: cobaya-grid-run %s --importance' % batchPath)
    print('for best-fit for importance sampled: '
          'cobaya-grid-run %s --importance_minimize' % batchPath)



# example Cobaya external likelihood package, here just a simple H0 prior

from setuptools import setup
import os

file_dir = os.path.abspath(os.path.dirname(__file__))
os.chdir(file_dir)

setup(name="test_package",
      version='1.0',
      description='Example external Cobaya likelihood package',
      zip_safe=True,  # set to false if you want to easily access bundled package data files
      packages=['test_package', 'test_package.sub_module', 'test_package.tests'],
      package_data={'test_package': ['*.yaml', "*.bibtex"],
                    'test_package.sub_module': ['*.yaml']},
      install_requires=['cobaya (>=2.0.5)'],
      test_suite='test_package.tests',
      )



import logging
import os
import numpy as np
import pytest
from cobaya import mpi, run, Theory, InputDict, PostDict, LoggedError, load_samples
from cobaya.conventions import Extension
from cobaya.component import ComponentNotFoundError
from cobaya.tools import deepcopy_where_possible
from cobaya.cosmo_input.convert_cosmomc import cosmomc_root_to_cobaya_info_dict
from cobaya.log import NoLogging
from .common import process_packages_path
from .conftest import install_test_wrapper

pytestmark = pytest.mark.mpi


def likelihood(_self):
    return -(_self.provider.get_param('sigma8') - 0.7) ** 2 / 0.1 ** 2


def likelihood2(sigma8, joint):
    return -(sigma8 - 0.75) ** 2 / 0.07 ** 2 + joint * 0


class ATheory(Theory):
    params = {'As': None, 'As100': {'derived': True}}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['derived']['As100'] = params_values_dict['As'] * 100


class BTheory(Theory):
    params = {'As100': None, 'As1000': {'derived': True}}

    def calculate(self, state, want_derived=True, **params_values_dict):
        state['derived']['As1000'] = params_values_dict['As100'] * 10


class CTheory(Theory):
    params = {'AsX': None, 'As1000': {'derived': True}}


info: InputDict = {"params": {
    "ombh2": 0.0245,
    "H0": 70,
    "ns": 0.965,
    "logA": {'prior': [1, 4], 'latex': r'\log(10^{10} A_\mathrm{s})', "proposal": 0.1},
    "As": {'value': lambda logA: 1e-10 * np.exp(logA), 'latex': r'A_\mathrm{s}'},
    "omegab": {'value': lambda ombh2, H0: ombh2 / (H0 / 100) ** 2,
               'latex': r'\Omega_\mathrm{b}'},
    "omegam": {'value': 0.3, 'latex': r'\Omega_\mathrm{m}', 'drop': True},
    "omch2": {'value': lambda omegam, omegab, mnu, H0: (omegam - omegab) * (
            H0 / 100) ** 2 - (mnu * (3.044 / 3) ** 0.75) / 94.0708,
              'latex': r'\Omega_\mathrm{c} h^2'},
    "tau": 0.05,
    "mnu": 0.0,
    "sigma8": None,
    "sigma82": {"derived": "lambda sigma8, omegam: sigma8 ** 2 * omegam"},
    "joint": {"derived": "lambda sigma8, omegam, As100: sigma8 ** 2 * omegam*As100"}},
    "likelihood": {
        "test_likelihood": {'external': likelihood,
                            'type': 'type1',
                            'requires': ['As', 'sigma8']}},
    'theory': {'test_theory': ATheory,
               'camb': {'stop_at_error': True,
                        'extra_args': {'num_massive_neutrinos': 1,
                                       'halofit_version': 'mead'}}},
    'sampler': {'mcmc': {'Rminus1_stop': 0.3}}
}


def test_cosmo_run_not_found():
    with NoLogging(logging.ERROR):
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["H0.perfect"] = None
        with pytest.raises(ComponentNotFoundError):
            run(inf)
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["none"] = None
        with pytest.raises(ComponentNotFoundError):
            run(inf)
        inf = deepcopy_where_possible(info)
        inf["likelihood"]["pandas.plotting.PlotAccessor"] = None
        with pytest.raises(LoggedError) as e:
            run(inf)
        assert "Failed to get defaults for component" in str(e)


@mpi.sync_errors
def test_cosmo_run_resume_post(tmpdir, skip_not_installed, packages_path=None):
    # only vary As, so fast chain. Chain does not need to converge (tested elsewhere).
    info['output'] = os.path.join(tmpdir, 'testchain')
    if packages_path:
        info["packages_path"] = process_packages_path(packages_path)
    install_test_wrapper(skip_not_installed, run, info, force=True)
    # note that continuing from files leads to text-file precision at read in, so a mix of
    # precision in the output SampleCollection returned from run
    run(info, resume=True, override={'sampler': {'mcmc': {'Rminus1_stop': 0.2}}})
    run(info, resume=True, override={'sampler': {'mcmc': {'Rminus1_stop': 0.15}},
                                     'theory': {'camb': {'stop_at_error': False}}},
        allow_changes=True)
    updated_info, sampler_init = run(
        info['output'] + '.updated' + Extension.dill, resume=True,
        override={'sampler': {'mcmc': {'Rminus1_stop': 0.05}}}
    )
    samp = sampler_init.samples(skip_samples=0.2, to_getdist=True)
    assert np.isclose(samp.mean('As100'), 100 * samp.mean('As'))

    # post-processing
    info_post: PostDict = {'add': {'params': {'h': None},
                                   "likelihood": {"test_likelihood2": likelihood2}},
                           'remove': {'likelihood': ["test_likelihood"]},
                           'suffix': 'testpost',
                           'skip': 0.2, 'thin': 4
                           }

    output_info, post_results = run(updated_info, override={'post': info_post},
                                    force=True)
    samp2 = post_results.samples(to_getdist=True)
    samp_test = samp.copy()
    samp_test.weighted_thin(4)
    sigma8 = samp_test.getParams().sigma8
    samp_test.reweightAddingLogLikes(-(sigma8 - 0.7) ** 2 / 0.1 ** 2
                                     + (sigma8 - 0.75) ** 2 / 0.07 ** 2)
    assert np.isclose(samp_test.mean('sigma8'), samp2.mean('sigma8'))

    # from getdist-format chain files
    root = os.path.join(tmpdir, 'getdist_format')
    if mpi.is_main_process():
        samp.saveChainsAsText(root)
    mpi.sync_processes()

    from_txt = dict(updated_info, output=root)
    post_from_text = dict(info_post, skip=0)  # getdist already skipped
    output_info, post_results_text = run(
        from_txt, override={'post': post_from_text}, force=True)
    samp_getdist = post_results_text.samples(to_getdist=True)
    assert not post_results_text["stats"]["points_removed"]
    assert samp2.numrows == samp_getdist.numrows
    assert np.isclose(samp2.mean('sigma8'), samp_getdist.mean('sigma8'))

    # again with inferred-inputs for params
    info_conv = cosmomc_root_to_cobaya_info_dict(root)
    # have to manually add consistent likelihoods if re-computing
    info_conv['likelihood'] = info['likelihood']
    info_conv['theory'] = info['theory']
    post_from_text = dict(info_post, skip=0, suffix='getdist2')  # getdist already skipped
    output_info, post_results_text = run(info_conv, override={'post': post_from_text},
                                         output=False)
    samp_getdist2 = post_results_text.samples(to_getdist=True)
    assert np.isclose(samp2.mean('sigma8'), samp_getdist2.mean('sigma8'))

    # from saved info, no output
    info_post['output'] = None
    output_info, post_results_from_saved = run(
        {'output': info['output'], 'post': info_post}, force=True)
    samp3 = post_results_from_saved.samples(to_getdist=True)
    assert np.isclose(samp3.mean("sigma8"), samp2.mean("sigma8"))
    assert np.isclose(samp3.mean("joint"), samp2.mean("joint"))
    samps4 = load_samples(info['output'] + '.post.testpost', to_getdist=True)
    assert np.isclose(samp3.mean("joint"), samps4.mean("joint"))

    # test recover original answer swapping likelihoods back
    info_revert = {'add': {'likelihood': info['likelihood']},
                   'remove': {'likelihood': ["test_likelihood2"]},
                   'suffix': 'revert',
                   'skip': 0, 'thin': 1,
                   'output': None
                   }
    output_info, post_products_revert = run({'output': info['output'] + '.post.testpost',
                                             'post': info_revert}, force=True)
    samp_revert = post_products_revert.samples(to_getdist=True)

    samp_thin = sampler_init.samples(skip_samples=0.2, to_getdist=True)
    samp_thin.weighted_thin(4)
    assert samp_thin.numrows == samp_revert.numrows + \
           post_products_revert["stats"]["points_removed"]
    if not post_products_revert["stats"]["points_removed"]:
        assert np.isclose(samp_revert.mean("sigma8"), samp_thin.mean("sigma8"))
    else:
        assert abs(samp_revert.mean("sigma8") - samp_thin.mean("sigma8")) < 0.01
    assert not post_products_revert["stats"]["points_removed"]

    # no remove
    info_post = {
        'add': {'params': {'h': None}, "likelihood": {"test_likelihood2": likelihood2}},
        'suffix': 'test2', 'skip': 0.2, 'thin': 4}
    output_info, post_results_no_remove = run(
        updated_info, override={'post': info_post}, force=True)
    samp2 = post_results_no_remove.samples(to_getdist=True)
    assert "chi2__type1" in samp2.paramNames.list()
    # check what has been saved to disk is consistent
    samps4 = load_samples(updated_info['output'] + '.post.test2', to_getdist=True)
    assert samp2.paramNames.list() == samps4.paramNames.list()
    assert np.isclose(samp2.mean("sigma8"), samps4.mean("sigma8"))

    # adding new theory derived
    info_post['add']['theory'] = {'new_param_theory': BTheory}
    output_info, post_results_derived = run(
        updated_info, override={'post': info_post}, output=False)
    samp3 = post_results_derived.samples(to_getdist=True)
    assert np.isclose(samp3.mean("sigma8"), samp2.mean("sigma8"))
    assert np.isclose(samp3.mean("As1000"), samp2.mean("As") * 1000)

    info_post['add']['theory'] = {'new_param_theory': CTheory}
    with pytest.raises(LoggedError) as e, NoLogging(logging.ERROR):
        run(updated_info, override={'post': info_post}, output=False)
    assert 'Parameter AsX no known value' in str(e)



import os
from cobaya.yaml import yaml_load_file
from cobaya import grid_tools
from cobaya.grid_tools import grid_create, grid_run, grid_converge, grid_tables, \
    grid_param_compare, grid_getdist, grid_list, grid_copy, grid_extract, grid_cleanup
from .common import stdout_check


def test_grid(tmpdir):

    f = os.path.join(tmpdir, 'grid')
    grid_create([f, os.path.join(os.path.dirname(__file__), 'simple_grid.py')])
    assert os.path.exists(os.path.join(f, 'base', 'like1_like2'))
    assert os.path.exists(os.path.join(f, 'input_files', 'base_a_2_like1_like2.yaml'))
    assert not os.path.exists(os.path.join(f, 'input_files', 'base_a_1_like1_like2.yaml'))

    info = yaml_load_file(
        os.path.join(f, 'input_files', 'base_a_1_a_2_like1_like2.yaml'))
    assert info['sampler']['mcmc']['max_samples'] == 100

    grid_run([f, '--noqueue', '2'])
    grid_run([f, '--noqueue', '--importance'])
    grid_getdist([f, '--burn_remove', '0.3'])
    assert os.path.exists(os.path.join(f, 'base_a_1_a_2',
                                       'like1', 'dist', 'base_a_1_a_2_like1.margestats'))
    assert os.path.exists(os.path.join(f, 'base_a_1_a_2', 'like1_like2',
                                       'base_a_1_a_2_like1_like2.post.cut.1.txt'))

    grid_run([f, '--noqueue', '--minimize', '--name', 'base_a_2_like1'])
    assert os.path.exists(os.path.join(f, 'base_a_2', 'like1', 'base_a_2_like1.minimum'))

    grid_run([f, '--noqueue', '--importance_minimize', '--name',
              'base_a_1_a_2_like1_like2.post.cut'])
    assert os.path.exists(os.path.join(f, 'base_a_1_a_2', 'like1_like2',
                                       'base_a_1_a_2_like1_like2.post.cut.minimum'))

    table_file = os.path.join(tmpdir, 'table')
    grid_tables([f, table_file, '--forpaper'])  # haven't installed latex in general
    assert os.path.exists(table_file + '.tex')

    grid_tables([f, table_file, '--limit', '1', '--forpaper',
                 '--param', 'a_2', '--data', 'like2'])
    with open(table_file + '.tex') as r:
        assert '68\\%' in r.read()

    with stdout_check("base_like1_like2 (main)"):
        grid_list(f)

    with stdout_check("10 dist files", "1 chain file"):
        grid_copy([f, os.path.join(tmpdir, 'test_grid.zip'), '--dist', '--chains',
                   '--remove_burn_fraction', '0.3', '--datatag', 'like1_like2'])

    with stdout_check("base_a_2_like1_like2 None"):
        grid_converge([f])

    with stdout_check("base_like1_like2"):
        grid_converge([f, '--checkpoint'])

    grid_param_compare([f, '--params', 'a_1', 'a_2', '--latex_filename', table_file])
    with open(table_file + '.tex') as r:
        assert 'a_2 &' in r.read()

    grid_extract([f, tmpdir, '.margestats', '--datatag', 'like1_like2'])
    assert os.path.exists(os.path.join(tmpdir, 'base_a_1_a_2_like1_like2.margestats'))

    with stdout_check("7 existing chains"):
        grid_create([f])

    with stdout_check("base_a_1_a_2_like1_like2"):
        grid_cleanup([f, '--confirm', '--data', 'like2'])

    with stdout_check("like2", match=False):
        grid_cleanup([f])


def test_cosmo_grid(tmpdir):
    test_name = 'base_w_planck_lowl_NPIPE_TTTEEE_lensing'
    f = os.path.join(tmpdir, 'grid')
    grid_create([f, os.path.join(os.path.dirname(__file__), 'test_cosmo_grid.yaml')])

    info = yaml_load_file(
        os.path.join(f, 'input_files', 'base_mnu_planck_lowl_NPIPE_TT.yaml'))
    assert info['theory']['camb']['extra_args']['num_massive_neutrinos'] == 3

    with stdout_check(test_name):
        grid_run([f, '--dryrun', '--job-template',
                  os.path.normpath(os.path.join(os.path.dirname(grid_tools.__file__),
                                                'script_templates/job_script_UGE'))])

    with stdout_check(test_name):
        grid_list(f)

    with stdout_check(test_name):
        grid_copy([f, os.path.join(tmpdir, 'grid_out')])

    with stdout_check("0 MB"):
        grid_cleanup([f])

    with stdout_check("Chains do not exist yet"):
        grid_getdist([f])

    with stdout_check("Chains do not", match=False):
        grid_getdist([f, '--exist'])



from .conftest import install_test_wrapper
import numpy as np
from cobaya.model import get_model
from cobaya.log import LoggedError, NoLogging

import pytest
import logging

cosmology_params = {
    'ombh2': 0.022, 'omch2': 0.12,
    'H0': 68,
    'ns': 0.96,
    'mnu': 0.06, 'nnu': 3.046
}


def _get_model(params, skip_not_installed):
    info = {
        'params': params,
        'likelihood':
            {'one':
                 {'requires':
                      {"Pk_grid":
                           {'k_max': 10, 'z': 0},
                       "As": None,
                       "sigma8": None
                       }
                  }
             },
        'theory': {'camb': {'stop_at_error': True,
                            'extra_args': {'num_massive_neutrinos': 1,
                                           'halofit_version': 'mead'}}}}
    return install_test_wrapper(skip_not_installed, get_model, info)


def test_CAMB_sigma8_input(skip_not_installed):
    power_params_s8 = {
        'sigma8': 0.78,
        'omegam': None,
        's8': {'derived': 'lambda sigma8, omegam: sigma8*np.sqrt(omegam/0.3)'},
        'As': None,
    }
    model_s8 = _get_model(
        {**cosmology_params, **power_params_s8},
        skip_not_installed)
    model_s8.loglike({})

    k, z, pk_s8 = model_s8.provider.get_Pk_grid()
    As_from_sigma8 = model_s8.provider.get_param("As")
    sigma8 = model_s8.provider.get_param("sigma8")

    model_as = _get_model(
        {**cosmology_params, "As": As_from_sigma8},
        skip_not_installed)
    model_as.loglike({})

    k, z, pk_as = model_as.provider.get_Pk_grid()

    assert np.isclose(sigma8, model_as.provider.get_param("sigma8"))
    assert np.allclose(pk_s8, pk_as)


def test_CAMB_As_and_sigma8_input_error(skip_not_installed):
    power_params_s8 = {
        'sigma8': 0.78,
        'omegam': None,
        's8': {'derived': 'lambda sigma8, omegam: sigma8*np.sqrt(omegam/0.3)'},
        'As': 2.1e-9,
    }

    with pytest.raises(LoggedError), NoLogging(logging.ERROR):
        model_s8 = _get_model(
            {**cosmology_params, **power_params_s8},
            skip_not_installed)
        model_s8.loglike({})



import os
import numpy as np
from cobaya.likelihoods.base_classes import SN


class PantheonPlus(SN):
    """
    Likelihood for Pantheon+ (without SH0ES) type Ia supernovae sample.

    Reference
    ---------
    https://arxiv.org/abs/2202.04077
    """

    def init_params(self, ini):
        self.twoscriptmfit = False
        data_file = os.path.normpath(os.path.join(self.path, ini.string("data_file")))
        self._read_data_file(data_file)
        self.covs = {}
        for name in ['mag']:
            self.log.debug('Reading covmat for: %s ' % name)
            self.covs[name] = self._read_covmat(
                os.path.join(self.path, ini.string('%s_covmat_file' % name)))
        self.alphabeta_covmat = False
        self.configure()
        self.inverse_covariance_matrix()
        if not self.use_abs_mag:
            self._marginalize_abs_mag()
        self.marginalize = False

    def _apply_mask(self, zmask):
        for col in self.cols:
            setattr(self, col, getattr(self, col)[zmask])
        for name, cov in self.covs.items():
            self.covs[name] = cov[np.ix_(zmask, zmask)]

    def configure(self):
        self._apply_mask(zmask=self.zcmb > 0.01)
        self.pre_vars = 0.  # diagonal component

    def _read_cols(self, data_file, file_cols, sep=None):
        self.log.debug('Reading %s' % data_file)
        with open(data_file, 'r') as f:
            lines = f.readlines()
            line = lines[0]
            if line.startswith('#'):
                line = line[1:]
            cols = [col.strip().lower() for col in line.split(sep)]
            assert cols[0].isalpha()
            indices = [cols.index(col) for col in file_cols]
            zeros = np.zeros(len(lines) - 1)
            for col in self.cols:
                setattr(self, col, zeros.astype(dtype='f8', copy=True))
            for ix, line in enumerate(lines[1:]):
                vals = [val.strip() for val in line.split(sep)]
                vals = [vals[i] for i in indices]
                for i, (col, val) in enumerate(zip(self.cols, vals)):
                    tmp = getattr(self, col)
                    tmp[ix] = np.asarray(val, dtype=tmp.dtype)
        self.nsn = ix + 1
        self.log.debug('Number of SN read: %s ' % self.nsn)

    def _read_data_file(self, data_file):
        file_cols = ['m_b_corr', 'zhd', 'zhel']
        self.cols = ['mag', 'zcmb', 'zhel']
        self._read_cols(data_file, file_cols)

    def _marginalize_abs_mag(self):
        deriv = np.ones_like(self.mag)[:, None]
        derivp = self.invcov.dot(deriv)
        fisher = deriv.T.dot(derivp)
        self.invcov = self.invcov - derivp.dot(np.linalg.solve(fisher, derivp.T))

    def alpha_beta_logp(self, lumdists, Mb=0., **kwargs):
        if self.use_abs_mag:
            estimated_scriptm = Mb + 25
        else:
            estimated_scriptm = 0.
        diffmag = self.mag - lumdists - estimated_scriptm
        return - diffmag.dot(self.invcov).dot(diffmag) / 2.



from .EE import EE


class EE_sroll2(EE):
    """
    Python translation of the Planck sroll2 EE likelihood
    https://web.fe.infn.it/~pagano/low_ell_datasets/sroll2/

    Calibration error is very small compared to EE uncertainty, but calibration can be
    used; otherwise taken to be 1
    """
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "planck_sroll2_lowE.zip",
                       "directory": "planck_sroll2_lowE_native"}

    _table_file_name = 'sroll2_prob_table.txt'

    @classmethod
    def get_bibtex(cls):
        return cls.get_associated_file_content('.bibtex')



import os
from cobaya.doc import doc_script
from cobaya.bib import bib_script
from .common import stdout_check


def test_doc():
    with stdout_check('planck_2018_lensing'):
        doc_script([])

    with stdout_check('DES_DzS1'):
        doc_script(['des_y1.shear'])

    with stdout_check(' [../base_classes/planck_calib, params_TT_CamSpec]'):
        doc_script(['planck_2018_highl_CamSpec.TT'])

    with stdout_check('lambda aksz, asz143'):
        doc_script(['planck_2018_highl_CamSpec.TT', '--expand'])


yaml = """
likelihood:
 des_y1.clustering:
 bao.generic:
 planck_2018_lensing.clik:
theory:
 camb:
sampler:
 minimize:
"""


def test_bib(tmpdir):
    with stdout_check('Neal:2005'):
        bib_script(['des_y1.shear', 'camb', 'mcmc'])

    f = os.path.join(tmpdir, 'input.yaml')
    with open(f, 'w', encoding='utf-8') as output:
        output.write(yaml)

    with stdout_check('Torrado:2020'):
        bib_script([f])



from scipy.stats import norm
from cobaya.likelihood import Likelihood


class TestLike(Likelihood):

    def initialize(self):
        self.norm = norm(loc=self.H0, scale=self.H0_std)

    def get_requirements(self):
        return {'H0': None}

    def logp(self, **params_values):
        H0_theory = self.provider.get_param("H0")
        return self.norm.logpdf(H0_theory)



from .pantheonplus import PantheonPlus


class PantheonPlusShoes(PantheonPlus):
    """
    Likelihood for Pantheon+ (with SH0ES) type Ia supernovae sample.

    Reference
    ---------
    https://arxiv.org/abs/2202.04077
    """

    def configure(self):
        self._apply_mask((self.zcmb > 0.01) | self.is_calibrator)
        self.pre_vars = 0.  # diagonal component

    def _read_data_file(self, data_file):
        file_cols = ['m_b_corr', 'zhd', 'zhel', 'is_calibrator', 'ceph_dist']
        self.cols = ['mag', 'zcmb', 'zhel', 'is_calibrator', 'ceph_dist']
        self._read_cols(data_file, file_cols)
        self.is_calibrator = self.is_calibrator.astype(dtype='?')

    def alpha_beta_logp(self, lumdists, Mb=0, **kwargs):
        if self.use_abs_mag:
            estimated_scriptm = Mb + 25
        else:
            estimated_scriptm = 0.

        # Use Cepheids host distances as theory
        lumdists[self.is_calibrator] = self.ceph_dist[self.is_calibrator] - 25.
        diffmag = self.mag - lumdists - estimated_scriptm
        return - diffmag.dot(self.invcov).dot(diffmag) / 2.



# Just plotting (loading on-the-fly)
# Notice that GetDist requires a full path when loading samples
import getdist.plots as gdplt
import os

folder, name = os.path.split(os.path.abspath(info["output"]))
gdplot = gdplt.get_subplot_plotter(chain_dir=folder)
gdplot.triangle_plot(name, ["a", "b"], filled=True)



import numpy as np
import matplotlib.pyplot as plt

_do_plot = False


def my_like(
        # Parameters that we may sample over (or not)
        noise_std_pixel=20,  # muK
        beam_FWHM=0.25,  # deg
        # Keyword through which the cobaya likelihood instance will be passed.
        _self=None):
    # Noise spectrum, beam-corrected
    healpix_Nside = 512
    pixel_area_rad = np.pi / (3 * healpix_Nside ** 2)
    weight_per_solid_angle = (noise_std_pixel ** 2 * pixel_area_rad) ** -1
    beam_sigma_rad = beam_FWHM / np.sqrt(8 * np.log(2)) * np.pi / 180.
    ells = np.arange(l_max + 1)
    Nl = np.exp((ells * beam_sigma_rad) ** 2) / weight_per_solid_angle
    # Cl of the map: data + noise
    Cl_map = Cl_est + Nl
    # Request the Cl from the provider
    Cl_theo = _self.provider.get_Cl(ell_factor=False, units="muK2")['tt'][:l_max + 1]
    Cl_map_theo = Cl_theo + Nl
    # Auxiliary plot
    if _do_plot:
        ell_factor = ells * (ells + 1) / (2 * np.pi)
        plt.figure()
        plt.plot(ells[2:], (Cl_theo * ell_factor)[2:], label=r'Theory $C_\ell$')
        plt.plot(ells[2:], (Cl_est * ell_factor)[2:], label=r'Estimated $C_\ell$',
                 ls="--")
        plt.plot(ells[2:], (Cl_map * ell_factor)[2:], label=r'Map $C_\ell$')
        plt.plot(ells[2:], (Nl * ell_factor)[2:], label='Noise')
        plt.legend()
        plt.ylim([0, 6000])
        plt.savefig(_plot_name)
        plt.close()
    # ----------------
    # Compute the log-likelihood
    V = Cl_map[2:] / Cl_map_theo[2:]
    logp = np.sum((2 * ells[2:] + 1) * (-V / 2 + 1 / 2. * np.log(V)))
    # Set our derived parameter
    derived = {'Map_Cl_at_500': Cl_map[500]}
    return logp, derived



"""
.. module:: cobaya.grid_tools.gridmanage

:Synopsis: tools for managing grids of chains (Cobaya version)
:Author: Antony Lewis (based on CosmoMC version of the same code)

"""

import os
import fnmatch
import subprocess
import time
import shutil
import zipfile
from datetime import datetime, timedelta
import getdist
from getdist import IniFile

from cobaya.conventions import Extension
from .batchjob_args import BatchArgs


# might be better as cobaya-grid command [options]

def run_and_wait(processes, commands=None, procs=1):
    if commands:
        processes.add(subprocess.Popen(commands))
    while len(processes) >= procs:
        time.sleep(.1)
        processes.difference_update(
            [p for p in processes if p.poll() is not None])


def grid_converge(args=None):
    opts = BatchArgs('Find chains which have failed or not converged, and show'
                     'Gelman-Rubin R-1 values for each run. Note need more than one'
                     'chain for getdist to calculate R-1.',
                     'cobaya-grid-converge', importance=True, converge=True)

    opts.parser.add_argument('--exist', action='store_true', help='chain must exist')
    opts.parser.add_argument('--checkpoint', action='store_true',
                             help='use R-1 stored in checkpoint files '
                                  '(rather than getdist output)')
    opts.parser.add_argument('--running', action='store_true',
                             help="only check running chains")
    opts.parser.add_argument('--not_running', action='store_true',
                             help="only check chains that are not running")
    opts.parser.add_argument('--stuck', action='store_true',
                             help="finds chains with big spread in the last update time")

    (batch, args) = opts.parseForBatch(args)

    not_exist = []
    converge = []

    if args.running:
        args.checkpoint = True

    if args.stuck:
        for job_item in opts.filteredBatchItems():
            if job_item.chainExists() and job_item.chainsDodgy():
                print('Chain stuck?...' + job_item.name)
    elif args.checkpoint:
        print('Convergence from checkpoint files...')
        for job_item in opts.filteredBatchItems():
            R, done = job_item.convergeStat()
            if R is not None and not done:
                if (not args.not_running or job_item.notRunning()) and (
                        not args.running or not job_item.notRunning()):
                    print('...', job_item.chainRoot, R)
                if args.running and job_item.chainExists() and job_item.chainsDodgy():
                    print('Chain stuck?...' + job_item.name)
    else:
        for job_item in opts.filteredBatchItems():
            if not job_item.chainExists():
                not_exist.append(job_item)
            elif args.converge == 0 or args.checkpoint or not \
                    job_item.hasConvergeBetterThan(args.converge, returnNotExist=True):
                if not args.not_running or job_item.notRunning():
                    converge.append(job_item)

        print('Checking batch (from last grid getdist output):')
        if not args.exist and len(not_exist) > 0:
            print('Not exist...')
            for job_item in not_exist:
                print('...', job_item.chainRoot)

        print('Converge check...')
        for job_item in converge:
            print('...', job_item.chainRoot, job_item.R())


def grid_getdist(args=None):

    opts = BatchArgs('Run getdist over the grid of models. '
                     'Use e.g. burn_remove=0.3 to remove 30% of the chain as burn in.',
                     'cobaya-grid-getdist', notExist=True)
    opts.parser.add_argument('--update_only', action='store_true',
                             help='only run if getdist on chains that have been updated '
                                  'since the last run')
    opts.parser.add_argument('--make_plots', action='store_true',
                             help='run generated script plot files to make PDFs')
    opts.parser.add_argument('--norun', action='store_true',
                             help='just make the .ini files, do not run getdist')
    opts.parser.add_argument('--burn_removed', action='store_true',
                             help="if burn in has already been removed from chains")
    opts.parser.add_argument('--burn_remove', type=float,
                             help="fraction of chain to remove as burn in "
                                  "(if not importance sampled or already done)")

    opts.parser.add_argument('--no_plots', action='store_true',
                             help="just make non-plot outputs "
                                  "(faster if using old plot_data)")
    opts.parser.add_argument('--delay', type=int,
                             help="run after delay of some number of seconds")
    opts.parser.add_argument('--procs', type=int, default=1,
                             help="number of getdist instances to run in parallel")
    opts.parser.add_argument('--base_ini', default=getdist.default_getdist_settings,
                             help="default getdist settings .ini file")
    opts.parser.add_argument('--command', default='getdist', help="program to run")
    opts.parser.add_argument('--exist', action='store_true',
                             help="Silently skip all chains that don't exist")

    (batch, args) = opts.parseForBatch(args)

    ini_dir = batch.batchPath + 'getdist' + os.sep
    os.makedirs(ini_dir, exist_ok=True)

    if args.delay:
        time.sleep(args.delay)
    processes = set()

    for job_item in opts.filteredBatchItems():
        ini = IniFile()
        ini.params['file_root'] = job_item.chainRoot
        ini.params['batch_path'] = job_item.batchPath
        os.makedirs(job_item.distPath, exist_ok=True)
        ini.params['out_dir'] = job_item.distPath
        if os.path.exists(args.base_ini):
            ini.defaults.append(args.base_ini)
        else:
            raise ValueError("base_ini file not found")
        if hasattr(batch, 'getdist_options'):
            ini.params.update(batch.getdist_options)
        tag = ''
        if job_item.isImportanceJob or args.burn_removed or job_item.isBurnRemoved():
            ini.params['ignore_rows'] = 0
        elif args.burn_remove is not None:
            ini.params['ignore_rows'] = args.burn_remove

        if job_item.isImportanceJob:
            ini.params['compare_num'] = 1
            ini.params['compare1'] = job_item.parent.chainRoot
        if args.no_plots:
            ini.params['no_plots'] = True
        if args.make_plots:
            ini.params['make_plots'] = True
        fname = ini_dir + job_item.name + tag + '.ini'
        ini.params.update(job_item.dist_settings)
        ini.saveFile(fname)
        if not args.norun and (not args.notexist or not job_item.getDistExists()) and (
                not args.update_only or job_item.getDistNeedsUpdate()):
            if job_item.chainExists():
                print("running: " + fname)
                run_and_wait(processes, [args.command] + [fname], args.procs)
            else:
                if not args.exist:
                    print("Chains do not exist yet: " + job_item.chainRoot)

    if args.procs > 1:
        run_and_wait(processes)


def grid_list(args=None):
    opts = BatchArgs('List items in a grid', 'cobaya-grid-list',
                     importance=True, converge=True, notExist=True)
    opts.parser.add_argument('--exists', action='store_true', help='chain must exist')
    opts.parser.add_argument('--normed', action='store_true', help='Output normed names')

    (batch, args) = opts.parseForBatch(args)
    items = opts.sortedParamtagDict(chainExist=args.exists)

    for paramtag, parambatch in items:
        for jobItem in parambatch:
            if hasattr(jobItem, 'group'):
                tag = '(%s)' % jobItem.group
            else:
                tag = ''
            if args.normed:
                print(jobItem.normed_name, tag)
            else:
                print(jobItem.name, tag)


def grid_cleanup(args=None):

    opts = BatchArgs('Delete failed chains, files etc. Nothing is actually delete'
                     'until you add --confirm, so you can check what you are doing first',
                     'cobaya-grid-cleanup', importance=True, converge=True)

    opts.parser.add_argument('--dist', action='store_true',
                             help="set to only affect getdist output files")
    opts.parser.add_argument('--ext', nargs='+', default=['*'],
                             help="file extensions to delete")
    opts.parser.add_argument('--empty', action='store_true')
    opts.parser.add_argument('--confirm', action='store_true')
    opts.parser.add_argument('--chainnum', default=None)

    (batch, args) = opts.parseForBatch(args)

    sizeMB = 0

    def fsizestr(_fname):
        nonlocal sizeMB
        sz = os.path.getsize(_fname) // 1024
        sizeMB += sz / 1024.
        if sz < 1024:
            return str(sz) + 'KB'
        if sz < 1024 * 1024:
            return str(sz // 1024) + 'MB'
        if sz < 1024 * 1024 * 1024:
            return str(sz // 1024 // 1024) + 'GB'

    if args.chainnum is not None:
        args.ext = ['.' + args.chainnum + '.' + ext for ext in args.ext]
    else:
        args.ext = ['.' + ext for ext in args.ext] + ['.*.' + ext for ext in args.ext]

    done = set()
    for jobItem in opts.filteredBatchItems():
        if (args.converge == 0 or not jobItem.hasConvergeBetterThan(
                args.converge, returnNotExist=True)) \
                and os.path.exists(jobItem.chainPath):
            dirs = [jobItem.chainPath]
            if args.dist:
                dirs = []
            if os.path.exists(jobItem.distPath):
                dirs += [jobItem.distPath]
            for adir in dirs:
                for f in sorted(os.listdir(adir)):
                    for ext in args.ext:
                        if fnmatch.fnmatch(f, jobItem.name + ext):
                            fname = adir + f
                            if fname not in done and os.path.exists(fname):
                                if not args.empty or os.path.getsize(fname) == 0:
                                    done.add(fname)
                                    print(fname, ' (' + fsizestr(fname) + ')')
                                    if args.confirm:
                                        os.remove(fname)

    print('Total size: %.3g MB' % sizeMB)
    if not args.confirm:
        print('Files not actually deleted: add --confirm to delete')


def grid_copy(args=None):

    opts = BatchArgs('copy or zip chains and optionally other files',
                     'cobaya-grid-copy', importance=True, converge=True)

    opts.parser.add_argument('target_dir', help="output root directory or zip file name")

    opts.parser.add_argument('--dist', action='store_true',
                             help="include getdist outputs")
    opts.parser.add_argument('--chains', action='store_true', help="include chain files")
    opts.parser.add_argument('--sym_link', action='store_true',
                             help="just make symbolic links to source directories")
    opts.parser.add_argument('--no_config', action='store_true',
                             help="don't copy grid config info")

    opts.parser.add_argument('--remove_burn_fraction', default=0.0, type=float,
                             help="fraction at start of chain to remove as burn in")

    opts.parser.add_argument('--file_extensions', nargs='+', default=['.*'],
                             help='extensions to include')
    opts.parser.add_argument('--skip_extensions', nargs='+',
                             default=['.locked', '.lock_err', Extension.progress,
                                      Extension.dill, Extension.checkpoint,
                                      '.corr', '.py', '.py_mcsamples', '.pysamples'])
    opts.parser.add_argument('--max_age_days', default=0.0, type=float,
                             help="only include files with date stamp "
                                  "at most max_age_days old")
    opts.parser.add_argument('--dryrun', action='store_true')
    opts.parser.add_argument('--verbose', action='store_true')
    opts.parser.add_argument('--zip', action='store_true',
                             help='make a zip file. Not needed if target_dir '
                                  'is a filename ending in .zip')

    (batch, args) = opts.parseForBatch(args)

    if args.target_dir.endswith('.zip'):
        args.zip = True
    if args.max_age_days:
        max_age = datetime.now() - timedelta(days=args.max_age_days)
    else:
        max_age = None

    sizeMB = 0

    if args.zip:
        zipper = zipfile.ZipFile(args.target_dir, 'w',
                                 compression=zipfile.ZIP_DEFLATED,
                                 allowZip64=True)
        target_dir = None
    else:
        zipper = None
        target_dir = os.path.abspath(args.target_dir) + os.sep
        os.makedirs(target_dir, exist_ok=True)

    if args.sym_link and (args.remove_burn_fraction or args.zip):
        raise Exception('option not compatible with --sym_link')

    def file_matches(_f, name):
        for ext in args.file_extensions:
            if fnmatch.fnmatch(_f, name + ext):
                for ext2 in args.skip_extensions:
                    if fnmatch.fnmatch(_f, name + ext2):
                        return False
                return True
        return False

    done = set()

    def do_copy(source, dest, _f, has_burn=False):
        nonlocal sizeMB
        if args.verbose:
            print(source + _f)
        frac = 1
        if not args.dryrun:
            if args.remove_burn_fraction and has_burn:
                lines = open(source + _f, encoding='utf-8-sig').readlines()
                lines = lines[:int((len(lines)) * args.remove_burn_fraction):]
                frac = 1 - args.remove_burn_fraction
            else:
                lines = None
            destf = dest + _f
            if destf in done:
                return
            done.add(destf)
            if args.zip:
                if lines:
                    zipper.writestr(destf, "".join(lines))
                else:
                    zipper.write(source + _f, destf)
            else:
                if lines:
                    open(target_dir + destf, 'w', encoding='utf-8').writelines(lines)
                else:
                    if args.sym_link:
                        if os.path.islink(target_dir + destf):
                            os.unlink(target_dir + destf)
                        os.symlink(os.path.realpath(source + _f), target_dir + destf)
                    else:
                        shutil.copyfile(source + _f, target_dir + destf)
        elif args.remove_burn_fraction and has_burn:
            frac = 1 - args.remove_burn_fraction
        sizeMB += os.path.getsize(source + _f) / 1024. ** 2 * frac

    def write_ini(iniName, _props):
        if args.dryrun:
            return
        if args.zip:
            zipper.writestr(outdir + iniName, "\n".join(_props.fileLines()))
        else:
            _props.saveFile(target_dir + outdir + iniName)

    if not args.no_config:
        config_path = os.path.join(batch.batchPath, 'config' + os.sep)
        if os.path.exists(config_path):
            if not args.dryrun and not args.zip:
                s = target_dir + 'config'
                os.makedirs(s, exist_ok=True)
            for f in os.listdir(config_path):
                do_copy(config_path, 'config' + os.sep, f)

    for jobItem in opts.filteredBatchItems():
        if args.converge == 0 or jobItem.hasConvergeBetterThan(args.converge):
            print(jobItem.name)
            chainfiles = 0
            infofiles = 0
            distfiles = 0
            doneProperties = False
            outdir = jobItem.relativePath
            if not args.zip:
                s1 = target_dir + outdir
                os.makedirs(s1, exist_ok=True)
            if args.chains and jobItem.chainExists() and (
                    not args.max_age_days or
                    datetime.fromtimestamp(jobItem.chainFileDate()) > max_age):
                i = 1
                while os.path.exists(jobItem.chainRoot + '.%d.txt' % i):
                    f = jobItem.name + '.%d.txt' % i
                    chainfiles += 1
                    do_copy(jobItem.chainPath, outdir, f, not jobItem.isImportanceJob)
                    i += 1
                if not jobItem.isImportanceJob and args.remove_burn_fraction:
                    props = jobItem.propertiesIni()
                    props.params['burn_removed'] = True
                    write_ini(jobItem.name + '.properties.ini', props)
                    doneProperties = True

            for f in os.listdir(jobItem.chainPath):
                if file_matches(f, jobItem.name):
                    if doneProperties and '.properties.ini' in f:
                        continue
                    if not args.max_age_days or datetime.fromtimestamp(
                            os.path.getmtime(jobItem.chainPath + f)) > max_age:
                        infofiles += 1
                        if args.verbose:
                            print(jobItem.chainPath + f)
                        do_copy(jobItem.chainPath, outdir, f)
            if args.dist and os.path.exists(jobItem.distPath):
                outdir += 'dist' + os.sep
                if not args.zip:
                    s2 = target_dir + outdir
                    os.makedirs(s2, exist_ok=True)
                for f in os.listdir(jobItem.distPath):
                    if file_matches(f, jobItem.name) \
                            and (not args.max_age_days or
                                 datetime.fromtimestamp(os.path.getmtime(
                                     jobItem.distPath + f)) > max_age):
                        distfiles += 1
                        do_copy(jobItem.distPath, outdir, f)
            print('... %d chain files, %d other files and %d dist files' % (
                chainfiles, infofiles, distfiles))

    if zipper:
        zipper.close()

    print('Total size: %u MB' % sizeMB)


def grid_extract(args=None):
    opts = BatchArgs('copy all files of a given type from all getdist '
                     'output directories in the grid', 'cobaya-grid-extract',
                     importance=True, converge=True)

    opts.parser.add_argument('target_dir')
    opts.parser.add_argument('file_extension', nargs='+')
    opts.parser.add_argument('--normalize_names', action='store_true',
                             help='replace actual name tags with normalized names')
    opts.parser.add_argument('--tag_replacements', nargs='+',
                             help="XX YY XX2 YY2 replaces name XX "
                                  "with YY, XX2 with YY2 etc.")

    (batch, args) = opts.parseForBatch(args)

    target_dir = os.path.abspath(args.target_dir) + os.sep
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    if args.tag_replacements is not None:
        replacements = {}
        for i, val in enumerate(args.tag_replacements[::2]):
            replacements[val] = args.tag_replacements[i * 2 + 1]
    else:
        replacements = None

    for ext in args.file_extension:
        if '.' not in ext:
            pattern = '.' + ext
        else:
            pattern = ext
        for jobItem in opts.filteredBatchItems():
            if os.path.exists(jobItem.distPath) and (
                    args.converge == 0 or jobItem.hasConvergeBetterThan(args.converge)):
                for f in os.listdir(jobItem.distPath):
                    if fnmatch.fnmatch(f, jobItem.name + pattern):
                        print(jobItem.distPath + f)
                        if args.normalize_names:
                            fout = jobItem.makeNormedName(replacements)[0] + \
                                   os.path.splitext(f)[1]
                        else:
                            fout = f
                        shutil.copyfile(jobItem.distPath + f, target_dir + fout)
